# Version 0.1
#
#  GeologicTimeScale is copyright 2005 Jeffrey G. Pittman.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA 02111-1307 USA

import sys, csv, string, cStringIO

from Products.GeologicTimeScale.utils import getrlcolor
from Products.GeologicTimeScale.utils import colorstr
from Products.GeologicTimeScale.utils import wordWrap
from Products.GeologicTimeScale.utils import rgb256

from Products.GeologicTimeScale.graphics import Line
from Products.GeologicTimeScale.graphics import DashedLine
from Products.GeologicTimeScale.graphics import Circle
from Products.GeologicTimeScale.graphics import Rectangle
from Products.GeologicTimeScale.graphics import Text
from Products.GeologicTimeScale.graphics import TransformedText
from Products.GeologicTimeScale.graphics import AlignedText
from Products.GeologicTimeScale.graphics import TransformedAndAlignedText

from reportlab.graphics.shapes import *
from reportlab.graphics.charts.textlabels import Label

# USGS colors:
usgs_colors = {'Geologic Time':'(180,180,180)',
               'Phanerozoic Eon':'(179,226,209)',
               'Cenozoic Era':'(255,255,0)',
               'Quaternary Period':'(255,255,77)',
               'Holocene Series':'(255,255,179)',
               'Pleistocene Series':'(255,235,98)',
               'Neogene System':'(253,204,138)',
               'Pliocene Series':'(254,235,172)',
               'Miocene Series':'(255,222,0)',
               'Paleogene System':'(255,179,0)',
               'Oligocene Series':'(234,198,114)',
               'Eocene Series':'(234,173,67)',
               'Paleocene Series':'(235,147,1)',
               'Mesozoic Era':'(127,173,81)',
               'Cretaceous System':'(127,195,28)',
               'Upper Cretaceous Series':'(222,241,151)',
               'Lower Cretaceous Series':'(179,223,127)',
               'Jurassic System':'(77,180,126)',
               'Upper Jurassic Series':'(204,235,197)',
               'Middle Jurassic Series':'(127,202,147)',
               'Lower Jurassic Series':'(102,192,146)',
               'Triassic System':'(103,195,183)',
               'Upper Triassic Series':'(204,236,225)',
               'Middle Triassic Series':'(153,215,190)',
               'Lower Triassic Series':'(103,179,159)',
               'Paleozoic Era':'(128,181,213)',
               'Permian System':'(103,198,221)',
               'Lopingian Series':'(179,227,238)',
               'Guadalupian Series':'(153,216,216)',
               'Cisuralian Series':'(128,206,201)',
               'Carboniferous System':'(153,189,218)',
               'Pennsylvanian Subsystem':'(104,159,202)',
               'Upper Pennsylvanian Series':'(104,159,202)',
               'Middle Pennsylvanian Series':'(104,159,202)',
               'Lower Pennsylvanian Series':'(104,159,202)',
               'Mississippian Subsystem':'(128,145,173)',
               'Upper Mississippian Series':'(128,145,173)',
               'Middle Mississippian Series':'(128,145,173)',
               'Lower Mississippian Series':'(128,145,173)',
               'Devonian System':'(153,153,201)',
               'Upper Devonian Series':'(203,189,220)',
               'Middle Devonian Series':'(153,131,190)',
               'Lower Devonian Series':'(128,125,186)',
               'Silurian System':'(177,114,182)',
               'Pridoli Series':'(233,199,226)',
               'Ludlow Series':'(202,167,209)',
               'Wenlock Series':'(177,137,179)',
               'Llandovery Series':'(152,88,168)',
               'Ordovician System':'(249,129,166)',
               'Upper Ordovician Series':'(251,180,189)',
               'Middle Ordovician Series':'(250,154,177)',
               'Lower Ordovician Series':'(230,125,164)',
               'Cambrian System':'(251,128,95)',
               'Upper Cambrian (Furongian) Series':'(253,205,184)',
               'Middle Cambrian Series':'(232,174,151)',
               'Lower Cambrian Series':'(231,124,114)',
               'Precambrian':'(178,134,83)',
               'Proterozoic Eon':'(204,216,145)',
               'Neoproterozoic Era':'(202,165,149)',
               'Neoproterozoic III System':'(234,216,188)',
               'Ediacaran System':'(234,216,188)',
               'Cryogenian System':'(220,171,170)',
               'Tonian System':'(203,164,108)',
               'Mesoproterozoic Era':'(221,194,136)',
               'Paleoproterozoic Era':'(179,178,94)',
               'Archean Eon':'(153,173,172)',
               'Neoarchean Era':'(203,205,200)',
               'Mesoarchean Era':'(221,194,136)',
               'Paleoarchean Era':'(153,151,145)',
               'Eoarchean Era':'(128,144,144)',
               'Hadean Eon':'(255,128,128)'}

def yUp(y):
    return 4567.2 - float(y)

class GeologicTimeScaleDrawingEngine:
    """
    The basic structure of SVG emitting methods is a modification of:
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/325823,
    whereby the strarray() method emits a list of lines that are
    added to a growing SVG file.

    SVG help at www.carto.net was also very useful.    
    """
    def __init__(self,
                 point_events,
                 range_events,
                 name="timescale.svg",
                 height=2250,
                 width=500,
                 x_offset = 10,
                 y_offset = 10,
                 end_time = 0.0,
                 start_time = 4567.2,
                 geologic_time_width = 20,
                 eon_width = 20,
                 era_width = 20,
                 system_width = 20,
                 period_width = 100,
                 series_width = 100,
                 stage_width = 100,
                 simple_version = 1):
        self.point_events = point_events
        self.range_events = range_events
        self.height = height
        self.width = width
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.geologic_time_width = geologic_time_width
        self.eon_width = eon_width
        self.era_width = era_width
        self.system_width = system_width
        self.period_width = period_width
        self.series_width = series_width
        self.stage_width = stage_width
        self.simple_version = simple_version
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.name = name
        self.subunits = []

        self.geologic_time = self.getGeologicTime()

        self.add(self.geologic_time)

        if simple_version:
            self.pruneForSimple()

        self.pruneForStartAndEnd(start_time, end_time)

    def add(self,subunit): 
        self.subunits.append(subunit)

    def rl_draw(self, drawing):
        self.rl_drawLeftBarscaleSection(drawing)
        self.rl_drawHeadings(drawing)
        for subunit in self.subunits: 
            subunit.rl_draw(drawing)            
        self.rl_drawRightBoundaryLabels(drawing)

    def getTimeScaleSection(self):
        var = []
        for subunit in self.subunits: 
            var += subunit.strarray()            
        return var

    def getOpenerSection(self):
        opener_section = \
          ["<?xml version=\"1.0\"?>\n",
           "<svg  id=\"timescale\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" onload=\"Init(evt)\" onmousemove=\"GetTrueCoords(evt); ShowTooltip(evt, true)\" onmouseout=\"ShowTooltip(evt, false)\" height=\"%d\" width=\"%d\">\n" % (self.height,self.width)]
        return opener_section
                      
    def getDefsSection(self):
        defs_section = ["    <defs>\n"]
        defs_section += self.getCSS()
        defs_section += self.getEcmascript()
        defs_section += ["    </defs>\n"]
        return defs_section
                    
    def write_ecmascript(self,filename=None):
        if filename:
            self.ecmascriptname = filename
        else:
            self.ecmascriptname = self.name + ".js"
        file = open(self.ecmascriptname,'w')
        file.writelines(self.getEcmascript())
        file.close()

    def getCSS(self):
        css = \
          ["\n",
           "        <style type=\"text/css\">\n",
           "            <![CDATA[\n",
           "                body {font-family:Arial,Helvetica,Geneva,sans-serif;}\n",
           "                .str0 {\n",
           "                    stroke:#000000;\n",
           "                    stroke-width:0.1\n",
           "                }\n",
           "            ]]>\n",
           "        </style>\n",
           "\n"]
        return css

    def getEcmascript(self):
        """
        Note: at one time there were three description lines in the mouseover.
        These are commented out. This could be useful to have more info
            in the tooltip.
        """
        ecmascript = \
              ["        <script language=\"Javascript\"><![CDATA[\n",
           "        // This script from http://svg-whiz.com/svg/Tooltip.svg\n",
           "        var SVGDocument = null;\n",
           "        var SVGRoot = null;\n",
           "        var SVGViewBox = null;\n",
           "        var svgns = 'http://www.w3.org/2000/svg';\n",
           "        var xlinkns = 'http://www.w3.org/1999/xlink';\n",
           "        var toolTip = null;\n",
           "        var TrueCoords = null;\n",
           "        var tipBox = null;\n",
           "        var tipText = null;\n",
           "        var tipTitle = null;\n",
           #"        var tipDesc1 = null;\n",
           #"        var tipDesc2 = null;\n",
           #"        var tipDesc3 = null;\n",
           "\n",
           "        var lastElement = null;\n",
           "        var titleText = '';\n",
           #"        var titleDesc1 = '';\n",
           #"        var titleDesc2 = '';\n",
           #"        var titleDesc3 = '';\n",
           "        function Init(evt)\n",
           "        {\n",
           "        SVGDocument = evt.target.ownerDocument;\n",
           "        SVGRoot = document.documentElement;\n",
           "        TrueCoords = SVGRoot.createSVGPoint();\n",
           "        \n",
           "        toolTip = SVGDocument.getElementById('ToolTip');\n",
           "        tipBox = SVGDocument.getElementById('TipBox');\n",
           "        tipText = SVGDocument.getElementById('TipText');\n",
           "        tipTitle = SVGDocument.getElementById('TipTitle');\n",
           #"        tipDesc1 = SVGDocument.getElementById('tipDesc1');\n",
           #"        tipDesc2 = SVGDocument.getElementById('tipDesc2');\n",
           #"        tipDesc3 = SVGDocument.getElementById('tipDesc3');\n",
           "        };\n",
           "\n",
           "        function GetTrueCoords(evt)\n",
           "        {\n",
           "           // find the current zoom level and pan setting, and adjust the reported\n",
           "           //    mouse position accordingly\n",
           "           var newScale = SVGRoot.currentScale;\n",
           "           var translation = SVGRoot.currentTranslate;\n",
           "           TrueCoords.x = (evt.clientX - translation.x)/newScale;\n",
           "           TrueCoords.y = (evt.clientY - translation.y)/newScale;\n",
           "        };\n",
           "\n",
           "        function ShowTooltip(evt, turnOn)\n",
           "        {\n",
           "           try\n",
           "           {\n",
           "              if (!evt || !turnOn)\n",
           "              {\n",
           "                 toolTip.setAttributeNS(null, 'display', 'none');\n",
           "              }\n",
           "              else\n",
           "              {\n",
           "                 var tipScale = 1/SVGRoot.currentScale;\n",
           "                 var textWidth = 0;\n",
           "                 var tspanWidth = 0;\n",
           "                 var boxHeight = 20;\n",
           "\n",
           "                 tipBox.setAttributeNS(null, 'transform', 'scale(' + tipScale + ',' + tipScale + ')' );\n",
           "                 tipText.setAttributeNS(null, 'transform', 'scale(' + tipScale + ',' + tipScale + ')' );\n",
           "\n",
           "                 var targetElement = evt.target;\n",
           "\n",
           "                 if ( lastElement != targetElement )\n",
           "                 {\n",
           "                    var targetTitle = targetElement.getElementsByTagName('title').item(0);\n",
           "                    if ( targetTitle )\n",
           "                    {\n",
           "                       titleText = targetTitle.firstChild.nodeValue;\n",
           "                       tipTitle.firstChild.nodeValue = titleText;\n",
           "                    }\n",
           "\n",
           #"                    var targetDesc1 = targetElement.getElementsByTagName('desc1').item(0);\n",
           #"                    if ( targetDesc1 )\n",
           #"                    {\n",
           #"                       titleDesc1 = targetDesc1.firstChild.nodeValue;\n",
           #"                       tipDesc1.firstChild.nodeValue = titleDesc1;\n",
           #"                    }\n",
           #"\n",
           #"                    var targetDesc2 = targetElement.getElementsByTagName('desc2').item(0);\n",
           #"                    if ( targetDesc2 )\n",
           #"                    {\n",
           #"                       titleDesc2 = targetDesc2.firstChild.nodeValue;\n",
           #"                       tipDesc2.firstChild.nodeValue = titleDesc2;\n",
           #"                    }\n",
           #"\n",
           #"                    var targetDesc3 = targetElement.getElementsByTagName('desc3').item(0);\n",
           #"                    if ( targetDesc3 )\n",
           #"                    {\n",
           #"                       titleDesc3 = targetDesc3.firstChild.nodeValue;\n",
           #"                       tipDesc3.firstChild.nodeValue = titleDesc3;\n",
           #"                    }\n",
           "                 }\n",
           "\n",
           "                 var xPos = TrueCoords.x + (10 * tipScale);\n",
           "                 var yPos = TrueCoords.y + (10 * tipScale);\n",
           "\n",
           "                 //return rectangle around object as SVGRect object\n",
           "                 var outline = tipText.getBBox();\n",
           "                 tipBox.setAttributeNS(null, 'width', Number(outline.width) + 10);\n",
           "                 tipBox.setAttributeNS(null, 'height', Number(outline.height) + 10);\n",
           "\n",
           "                 toolTip.setAttributeNS(null, 'transform', 'translate(' + xPos + ',' + yPos + ')');\n",
           "                 toolTip.setAttributeNS(null, 'display', 'inline');\n",
           "              }\n",
           "           }\n",
           "           catch(er){}\n",
           "         };\n",
           "        ]]></script>\n"]
        return ecmascript

    def rl_drawLeftBarscaleSection(self, drawing):
        width = 10
        height = 100
        color1 = getrlcolor(133,133,122)
        color2 = getrlcolor(200,200,200)
        fontsize = 10
        fontcolor = getrlcolor(0,0,0)
        rect_x = self.x_offset - width
        text_x = rect_x - (width / 2.0) - 3
        y = self.y_offset
        var = []
        alternator = 1
        color = color1
        for millenium in range(45):
            if alternator:
                color = color1
                alternator = 0
            else:
                color = color2
                alternator = 1
            drawing.add(Rect(rect_x, yUp(y+height), width, height, fillColor=color))
            label = Label()
            label.setOrigin(text_x, yUp(y))
            label.angle = 90
            label.fillColor = fontcolor
            label.fontSize = fontsize
            label.setText("%d" % (millenium*100))
            drawing.add(label)
            y += 100

        height = 68
        if alternator:
            color = color1
        else:
            color = color2
        drawing.add(Rect(rect_x, yUp(y+height), width, height, fillColor=color))
        label = Label()
        label.setOrigin(text_x, yUp(y))
        label.angle = 90
        label.fillColor = fontcolor
        label.fontSize = fontsize
        label.setText("4500")
        drawing.add(label)

        y += 67
        label = Label()
        label.setOrigin(text_x, yUp(y))
        label.angle = 90
        label.fillColor = fontcolor
        label.fontSize = fontsize
        label.setText("4567.2")
        drawing.add(label)

    def getLeftBarscaleSection(self):
        width = 10
        height = 100
        color1 = (133,133,122)
        color2 = (200,200,200)
        fontsize = 10
        fontcolor = "#000"
        rect_x = self.x_offset - width
        text_x = rect_x - (fontsize/3)
        y = self.y_offset
        var = []
        alternator = 1
        var += ["    <g id=\"bar_scale\">\n"]
        for millenium in range(45):
            if alternator:
                color = color1
                alternator = 0
            else:
                color = color2
                alternator = 1
            var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\"\n" %\
                    (rect_x,y,height),
                    "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                    (width,colorstr(color))]
            var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,y),
                    "            <g transform=\"rotate(-90)\">\n",
                    "                <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">%d</text>\n" %\
                    (fontsize,fontcolor,millenium*100),
                    "            </g>\n",
                    "        </g>\n"]
            y += 100
        height = 68
        if alternator:
            color = color1
        else:
            color = color2
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\"\n" %\
                (rect_x,y,height),
                "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (width,colorstr(color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">4500</text>\n" %\
                (fontsize,fontcolor),
                "            </g>\n",
                "        </g>\n"]
        y += 67
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">4567.2</text>\n" %\
                (fontsize,fontcolor),
                "            </g>\n",
                "        </g>\n",
                "    </g>\n"]
        return var

    def getPointEventsSection(self):
        var = []
        var += ["    <g id=\"point_events\">\n"]

        for event in self.point_events:
            if event.symbol_type == 'ellipse':
                var += ["        <ellipse cx=\"%d\" cy=\"%d\" rx=\"%d\" ry=\"%d\" " %\
                        (event.x_position,self.y_offset+event.time,event.symbol_width,event.symbol_height),
                        "            style=\"stroke-width:0.2;fill:%s;\"></ellipse>\n" %\
                        (colorstr(event.color))]
            elif event.symbol_type == 'rect':
                var += ["        <rect x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" " %\
                        (event.x_position,self.y_offset+event.time,event.symbol_width,event.symbol_height),
                        "            style=\"stroke-width:0.2;fill:%s;\"></rect>\n" %\
                        (colorstr(event.color))]

            origin = (event.x_position, self.y_offset + event.time)
            textAnchor = 'start'

            if event.label_angle == 0:
                if event.label_placement == 'c':
                    origin = (event.x_position+(event.symbol_width/2.0), self.y_offset+event.time+(event.symbol_width/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    origin = (event.x_position-4, self.y_offset+event.time+(event.symbol_width/2.0))
                    textAnchor = 'end'
                elif event.label_placement== 'r': 
                    origin = (event.x_position+event.symbol_width+4, self.y_offset+event.time+(event.symbol_width/2.0))
                    textAnchor = 'start'
                elif event.label_placement== 't': 
                    origin = (event.x_position+(event.symbol_width/2.0), self.y_offset+event.time-4)
                    textAnchor = 'middle'
                elif event.label_placement== 'b': 
                    origin = (event.x_position+(event.symbol_width/2.0), self.y_offset+event.time+4)
                    textAnchor = 'middle'
            else: # assume label_angle = 90
                if event.label_placement == 'c':
                    origin = (event.x_position+(event.symbol_width/2.0), self.y_offset+event.time+(event.symbol_width/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    origin = (event.x_position-4, self.y_offset+event.time+(event.symbol_width/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 'r': 
                    origin = (event.x_position+event.symbol_width+4, self.y_offset+event.time+(event.symbol_width/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 't': 
                    origin = (event.x_position+(event.symbol_width/2.0), self.y_offset+event.time-4)
                    textAnchor = 'start'
                elif event.label_placement== 'b': 
                    origin = (event.x_position+(event.symbol_width/2.0), self.y_offset+event.time+4)
                    textAnchor = 'end'

            label = TransformedAndAlignedText(origin,
                                              "%s" % event.text,
                                              rgb256(event.color),
                                              event.font_size,
                                              event.label_angle,
                                              textAnchor)
            var += label.strarray()

        var += ["    </g>\n"]
        return var

    def getRangeEventsSection(self):
        var = []
        var += ["    <g id=\"range_events\">\n"]
        for event in self.range_events:
            var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\" " %\
                    (event.x_position,self.y_offset+event.end_time,(event.start_time-event.end_time)),
                    "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"></rect>\n" %\
                    (event.bar_width,colorstr(event.color))]

            origin = (event.x_position, self.y_offset + event.end_time)
            textAnchor = 'start'

            if event.label_angle == 0:
                if event.label_placement == 'c':
                    origin = (event.x_position+(event.bar_width/2.0), self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    origin = (event.x_position-4, self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0))
                    textAnchor = 'end'
                elif event.label_placement== 'r': 
                    origin = (event.x_position+event.bar_width+4, self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0))
                    textAnchor = 'start'
                elif event.label_placement== 't': 
                    origin = (event.x_position+(event.bar_width/2.0), self.y_offset+event.end_time-2)
                    textAnchor = 'middle'
                elif event.label_placement== 'b': 
                    origin = (event.x_position+(event.bar_width/2.0), self.y_offset+event.start_time+2)
                    textAnchor = 'middle'
            else: # assume label_angle = 90
                if event.label_placement == 'c':
                    origin = (event.x_position+(event.bar_width/2.0), self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    origin = (event.x_position-4, self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 'r': 
                    origin = (event.x_position+event.bar_width+4, self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0))
                    textAnchor = 'middle'
                elif event.label_placement== 't': 
                    origin = (event.x_position+(event.bar_width/2.0), self.y_offset+event.end_time-2)
                    textAnchor = 'start'
                elif event.label_placement== 'b': 
                    origin = (event.x_position+(event.bar_width/2.0), self.y_offset+event.start_time+2)
                    textAnchor = 'end'

            label = TransformedAndAlignedText(origin,
                                              "%s" % event.text,
                                              rgb256(event.color),
                                              event.font_size,
                                              -1.0*event.label_angle,
                                              textAnchor)
            var += label.strarray()

        var += ["    </g>\n"]
        return var

    def rl_drawRightBoundaryLabels(self, drawing):
        root = self.subunits[0]
        var = []
        eons = []
        self.getTimeUnits(eons,'Eon')
        for eon in eons:
            eon.rl_drawRightBoundaryLabels(var,'System',root,drawing)
        return var

    def getRightBoundaryLabelsSection(self):
        root = self.subunits[0]
        var = []
        eons = []
        self.getTimeUnits(eons,'Eon')
        for eon in eons:
            eon.getRightBoundaryLabels(var,'System',root)
        return var

    def rl_drawLabel(self, drawing, text_x, text_y, angle, boxAnchor, fontSize, fontColor, text):
        label = Label()
        label.setOrigin(text_x, yUp(text_y))
        label.angle = angle
        label.boxAnchor = boxAnchor
        label.fontSize = fontSize
        label.fillColor = getrlcolor(fontColor[0],fontColor[1],fontColor[2])
        label.setText(text)
        drawing.add(label)

    def rl_drawHeadings(self, drawing):
        angle = 90
        header_label_height = 50
        header_label_color = (133,233,222)
        header_label_fontsize = 10
        header_label_fontcolor = (0,0,0)
        rect_x = self.x_offset + self.geologic_time_width
        text_x = rect_x + (self.eon_width/2)
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2)
        drawing.add(Rect(rect_x,yUp(self.y_offset),self.eon_width,header_label_height,fillColor=getrlcolor(header_label_color[0],header_label_color[1],header_label_color[2])))
        self.rl_drawLabel(drawing, text_x, text_y, 90, 'c', header_label_fontsize, header_label_fontcolor, 'EON')

        rect_x += self.eon_width
        text_x = rect_x + (self.era_width/2)
        drawing.add(Rect(rect_x,yUp(self.y_offset),self.era_width,header_label_height,fillColor=getrlcolor(header_label_color[0],header_label_color[1],header_label_color[2])))
        self.rl_drawLabel(drawing, text_x, text_y, 90, 'c', header_label_fontsize, header_label_fontcolor, 'ERA')
        
        if self.simple_version:
            header_label_height /= 2
            rect_x += self.era_width
            rect_y = self.y_offset-header_label_height
            text_x = rect_x + (self.system_width/2)
            text_y = rect_y + (header_label_height/2)
            angle = 0
        else:
            rect_x += self.era_width
            text_x = rect_x + (self.system_width/2)
            angle = 90
        drawing.add(Rect(rect_x,yUp(self.y_offset),self.system_width,header_label_height,fillColor=getrlcolor(header_label_color[0],header_label_color[1],header_label_color[2])))
        self.rl_drawLabel(drawing, text_x, text_y, angle, 'c', header_label_fontsize, header_label_fontcolor, 'PERIOD')

        header_label_height /= 2
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2)
        rect_x += self.system_width
        text_x = rect_x + (self.series_width/2)
        drawing.add(Rect(rect_x,yUp(self.y_offset),self.series_width,header_label_height,fillColor=getrlcolor(header_label_color[0],header_label_color[1],header_label_color[2])))
        self.rl_drawLabel(drawing, text_x, text_y, 0, 'c', header_label_fontsize, header_label_fontcolor, 'EPOCH')

        if not self.simple_version:
            rect_x += self.series_width
            text_x = rect_x + (self.stage_width/2)
            drawing.add(Rect(rect_x,yUp(self.y_offset),self.stage_width,header_label_height,fillColor=getrlcolor(header_label_color[0],header_label_color[1],header_label_color[2])))
            self.rl_drawLabel(drawing, text_x, text_y, 0, 'c', header_label_fontsize, header_label_fontcolor, 'STAGE')

    def getHeaderLabelSection(self):
        header_label_height = 50
        header_label_color = (133,233,222)
        header_label_fontsize = 10
        header_label_fontcolor = "#000"
        rect_x = self.x_offset + self.geologic_time_width
        text_x = rect_x + (self.eon_width/2) + (header_label_fontsize/3)
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2) + (header_label_fontsize/3)
        var  = ["<g id=\"mainlayer\" style=\"fill-opacity:1.0; stroke:black;stroke-width:1;\">\n"]
        var += ["    <g id=\"header_labels\">\n"]
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\" width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
               (rect_x,rect_y,header_label_height,self.eon_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">EON</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "            </g></g>\n"]
        rect_x += self.eon_width
        text_x = rect_x + (self.era_width/2) + (header_label_fontsize/3)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\" " %\
                (rect_x,rect_y,header_label_height),
                "width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.era_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">ERA</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "            </g></g>\n"]
        rect_x += self.era_width
        text_x = rect_x + (self.system_width/2) + (header_label_fontsize/3)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\" " %\
                (rect_x,rect_y,header_label_height),
                "width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.system_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">SYSTEM</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "            </g></g>\n"]
        header_label_height /= 2
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2) + (header_label_fontsize/3)
        rect_x += self.system_width
        text_x = rect_x + (self.series_width/2)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\" " %\
                (rect_x,rect_y,header_label_height),
                "width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.series_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">SERIES</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "        </g>\n"]
        rect_x += self.series_width
        text_x = rect_x + (self.stage_width/2)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\" " %\
                (rect_x,rect_y,header_label_height),
                "width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.stage_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <text class=\"str0\" style=\"fill:black;font-size:%d;fill:%s;text-anchor:middle\">STAGE</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "        </g>\n",
                "    </g>\n"]
        return var
    
    def getSimpleHeaderLabelSection(self):
        header_label_height = 50
        header_label_color = (133,233,222)
        header_label_fontsize = 10
        header_label_fontcolor = "#000"
        rect_x = self.x_offset + self.geologic_time_width
        text_x = rect_x + (self.eon_width/2) + (header_label_fontsize/3)
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2) + (header_label_fontsize/3)
        var =  ["    <g id=\"header_labels\">\n"]
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\"\n" %\
               (rect_x,rect_y,header_label_height),
               "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
               (self.eon_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:#000;font-size:%d;font-color:%s;text-anchor:middle\">EON</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "            </g></g>\n"]
        rect_x += self.eon_width
        text_x = rect_x + (self.era_width/2) + (header_label_fontsize/3)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\"\n" %\
                (rect_x,rect_y,header_label_height),
                "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.era_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <g transform=\"rotate(-90)\">\n",
                "                <text class=\"str0\" style=\"fill:#000;font-size:%d;font-color:%s;text-anchor:middle\">ERA</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "            </g></g>\n"]
        header_label_height /= 2
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2) + (header_label_fontsize/3)
        rect_x += self.era_width
        text_x = rect_x + (self.series_width/2)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\"\n" %\
                (rect_x,rect_y,header_label_height),
                "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.system_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "                <text class=\"str0\" style=\"fill:#000;font-size:%d;font-color:%s;text-anchor:middle\">SYSTEM</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "        </g>\n"]
        rect_y = self.y_offset-header_label_height
        text_y = rect_y + (header_label_height/2) + (header_label_fontsize/3)
        rect_x += self.system_width
        text_x = rect_x + (self.series_width/2)
        var += ["        <rect x=\"%d\" y=\"%d\" height=\"%d\"\n" %\
                (rect_x,rect_y,header_label_height),
                "            width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"/>\n" %\
                (self.series_width,colorstr(header_label_color))]
        var += ["        <g transform=\"translate(%d,%d)\">\n" % (text_x,text_y),
                "            <text class=\"str0\" style=\"fill:#000;font-size:%d;font-color:%s;text-anchor:middle\">SERIES</text>\n" %\
                (header_label_fontsize,header_label_fontcolor),
                "        </g>\n",
                "    </g>\n"]
        return var
                
    def getBottomPopupSection(self):
        bottom_svg = \
          ["    <g id='ToolTip' opacity='0.8' display='none' pointer-events='none'>\n",
           "        <rect id='TipBox' x='0' y='5' width='88' height='20' rx='2' ry='2' fill='white' stroke='black'/>\n",
           "        <text id='TipText' x='5' y='20' font-family='Arial' font-size='12'>\n",
           "            <tspan id='TipTitle' x='5' font-weight='bold'> </tspan>\n",
           #"         <tspan id='tipDesc1' x='5' dy='1.2em' fill='blue'> </tspan>\n",
           #"         <tspan id='tipDesc2' x='5' dy='1.2em' fill='blue'> </tspan>\n",
           #"         <tspan id='tipDesc3' x='5' dy='1.2em' fill='blue'> </tspan>\n",
           "        </text>\n",
           "    </g>\n"]
        return bottom_svg
    
    def getCloserSection(self):
        closer_section = ["</svg>\n"]
        return closer_section
        
    def getSVG(self):
        svg_code = self.getOpenerSection()
        svg_code += self.getDefsSection()
        if self.simple_version:
            svg_code += self.getSimpleHeaderLabelSection()
        else:
            svg_code += self.getHeaderLabelSection()
        svg_code += self.getTimeScaleSection()
        svg_code += self.getLeftBarscaleSection()
        svg_code += self.getRightBoundaryLabelsSection()
        if self.simple_version:
            svg_code += self.getRangeEventsSection()
            svg_code += self.getPointEventsSection()
        svg_code += self.getBottomPopupSection()
        svg_code += self.getCloserSection()
        return svg_code

    def getTimeUnits(self,time_units,target_type):
        for subunit in self.subunits:
            subunit.getTimeUnits(time_units,target_type)

    def write_html(self,filename=None):
        if filename:
            self.htmlname = filename
        else:
            self.htmlname = self.name + ".html"
        file = open(self.htmlname,'w')
        file.writelines(self.getHTML(self.name))
        file.close()

    def _getHTMLForContainingImageMap(self, var, unit):
        if len(unit.subunits) > 0:
            for subunit in unit.subunits[:]:
                var += subunit.imageMapTag()
                self._getHTMLForContainingImageMap(var, subunit)
            
    def getHTMLForContainingImageMap(self, name):        
        var = ["<body>\n",
               "<map name=\"drawing.map\">\n"]

        # imagemap coordinates are like SVG coordinates, from the upper
        # left corder of the image, so use yUp as with SVG

        self._getHTMLForContainingImageMap(var, self.geologic_time)

        # y_offset is twice in the calls below because it was added 
        # into the rects already, which should probably be fixed.

        for point_event in self.point_events:
            var += point_event.imageMapTag(self.y_offset * 2)

        for range_event in self.range_events:
            var += range_event.imageMapTag(self.y_offset * 2)

        var += ["</map>\n",
                "<img usemap=\"drawing.map\" src=\"drawing.png\">\n",
                "</body>\n"]

        return var

    def getHTMLForContainingSVG(self, name):
        var = ["<body>\n",
               "<H1>Instructions below don't apply to all browsers.</H1>\n",
               "<ul>\n",
               " <li><b>Zooming</b>: use the CTRL-Key (or Apple-Key) to drag a rectangle and zoom in.\n",
               " <br />Use the right mouse button (CTRL-Key plus mouse click on the Mac) to get a context menu, <br />where you can choose step-by-step zooming and can go back to the original extent.</li>\n",
               " <li><b>Zoom Out</b>: like above, but with SHIFT-Key pressed.</li>\n", 
               " <li><b>Panning</b>: Press the ALT-Key and move the mouse cursor while a hand-icon appears.</li>\n",
               " <li><b>Anti-Aliasing</b>: May be turned on and off with the help of the context menu.  (Higher Quality).</li>\n",
               "</ul>\n",
               "<h2>SVG Phylogenetic Tree</h2>\n",
               "<embed width=\"100%%\" height=\"660px\" src=\"%s.svg\" type=\"image/svg+xml\" />\n" % name, 
               "<br>\n",
               "<form name=\"hilite_form\" action=\"\">\n",
               "  <table border=\"0\" cellpadding=\"0\" cellspacing=\"2\">\n",
               "    <tr>\n",
               "      <td colspan=\"2\">Visible Layers\n",
               "      </td>\n",
               "    </tr>\n",
               "    <tr>\n",
               "      <td><input type=\"checkbox\" value=\"\" onclick=\"hilite_elem(this,'bar_scale')\">&nbsp;Bar Scale</TD>\n",
               "      <td><input type=\"checkbox\" value=\"\" onclick=\"hilite_elem(this,'header_labels')\">&nbsp;Header Labels</TD>\n",
               "    </tr>\n",
               "   </table>\n",
               "   <script type=\"text/javascript\">\n",
               "     // Make sure all checkboxes are checked whenever the page is reloaded in the browser.\n",
               "     for (var i = 0;  i < document.hilite_form.elements.length;  i++)\n",
               "       if (document.hilite_form.elements[i].type == 'checkbox'){\n",
               "         document.hilite_form.elements[i].checked = true;\n",
               "       }  \n",
               "   </script>\n",
               "</form>\n",
               "</body>\n"]
        return var

    def _getHTMLForListing(self, var, unit):
        if len(unit.subunits) > 0:
            for subunit in unit.subunits[:]:
                var += subunit.htmlForListing()
                self._getHTMLForListing(var, subunit)
            
    def getHTMLForListing(self):
        var = ["<body>\n"]

        var += "<H1>%s</H1>\n" % ('Time Units')

        self._getHTMLForListing(var, self.geologic_time)

        for point_event in self.point_events:
            var += point_event.htmlForListing()

        for range_event in self.range_events:
            var += range_event.htmlForListing()

        var += "</body>\n" 
        
        return var 
        
    def getUnitColor(self, name):
        if usgs_colors.has_key(name):
            return usgs_colors[name]
        else:
            return None
        
    def getExistingUnit(self, name, high_level_units):
        for hlu in high_level_units:
            if hlu.isUnit(name):
                begin = hlu.getBegin()
                begin_error = hlu.getBeginError()
                if begin < 0.0:
                    return hlu
        return None

    def pruneForSimple(self):
        self._pruneForSimple(self.geologic_time)

    def _pruneForSimple(self, unit):
        if len(unit.subunits) > 0:
            for subunit in unit.subunits[:]:
                unit_name = subunit.name.split(' ')[0]
                unit_name = unit_name.strip()
                unit_type = subunit.name.split(' ')[-1]
                unit_type = unit_type.strip()
                if unit_type in ["Subsystem","Series","Subseries","Stage","Substage"] and unit_name not in ["Paleocene","Eocene","Oligocene","Miocene","Pliocene","Pleistocene","Holocene"]:
                    unit.subunits.remove(subunit)
                else:
                    self._pruneForSimple(subunit)

    def pruneForStartAndEnd(self,timescale_begin, timescale_end):
        self._pruneForStartAndEnd(self.geologic_time, timescale_begin, timescale_end)

    def _pruneForStartAndEnd(self, unit, timescale_begin, timescale_end):
        if unit.begin > timescale_begin:
            unit.begin = timescale_begin
        if unit.end < timescale_end:
            unit.end = timescale_end
        if len(unit.subunits) > 0:
            for subunit in unit.subunits[:]:
                if subunit.begin <= timescale_end or subunit.end >= timescale_begin:
                    unit.subunits.remove(subunit)
                else:
                    self._pruneForStartAndEnd(subunit, timescale_begin, timescale_end)

    def rl_drawPointEvents(self, drawing):
        for event in self.point_events:
            if event.symbol_type == 'ellipse':
                drawing.add(Ellipse(cx        = event.x_position,
                                    cy        = yUp(self.y_offset+event.time-event.symbol_height/2.0),
                                    rx        = event.symbol_width,
                                    ry        = event.symbol_height,
                                    fillColor = getrlcolor(event.color[0],event.color[1],event.color[2])))
            elif event.symbol_type == 'rect':
                drawing.add(Rect(x         = event.x_position,
                                 y         = yUp(self.y_offset+event.time-event.symbol_height/2.0),
                                 width     = event.symbol_width,
                                 height    = event.symbol_height,
                                 fillColor = getrlcolor(event.color[0],event.color[1],event.color[2])))
            label = Label()
            label.angle = event.label_angle
            label.setText(event.text)
            #label.fontName = 'Times-Roman'
            label.fontSize = event.font_size
            if label.angle == 0:
                if event.label_placement == 'c':
                    label.setOrigin(event.x_position+(event.symbol_width/2.0), yUp(self.y_offset+event.time+event.symbol_height/2.0))
                    label.boxAnchor = 'c'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    label.setOrigin(event.x_position-4, yUp(self.y_offset+event.time+event.symbol_height/2.0))
                    label.boxAnchor = 'e'
                    label.textAnchor = 'end'
                elif event.label_placement== 'r': 
                    label.setOrigin(event.x_position+event.symbol_width+4, yUp(self.y_offset+event.time))
                    label.boxAnchor = 'w'
                    label.textAnchor = 'start'
                elif event.label_placement== 't': 
                    label.setOrigin(event.x_position+(event.symbol_width/2.0), yUp(self.y_offset+event.time-4))
                    label.boxAnchor = 's'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'b': 
                    label.setOrigin(event.x_position+(event.symbol_width/2.0), yUp(self.y_offset+event.time+4))
                    label.boxAnchor = 'n'
                    label.textAnchor = 'middle'
            else: # assume label_angle = 90, and hope it works for other angles well enough
                if event.label_placement == 'c':
                    label.setOrigin(event.x_position+(event.symbol_width/2.0), yUp(self.y_offset+event.time+event.symbol_height/2.0))
                    label.boxAnchor = 'c'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    label.setOrigin(event.x_position-4, yUp(self.y_offset+event.time+event.symbol_height/2.0))
                    label.boxAnchor = 's'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'r': 
                    label.setOrigin(event.x_position+event.symbol_width+4, yUp(self.y_offset+event.time))
                    label.boxAnchor = 'n'
                    label.textAnchor = 'middle'
                elif event.label_placement== 't': 
                    label.setOrigin(event.x_position+(event.symbol_width/2.0), yUp(self.y_offset+event.time-4))
                    label.boxAnchor = 'w'
                    label.textAnchor = 'start'
                elif event.label_placement== 'b': 
                    label.setOrigin(event.x_position+(event.symbol_width/2.0), yUp(self.y_offset+event.time+4))
                    label.boxAnchor = 'e'
                    label.textAnchor = 'end'
            drawing.add(label)

    def rl_drawRangeEvents(self, drawing):
        for event in self.range_events:
            drawing.add(Rect(x         = event.x_position,
                             y         = yUp(self.y_offset+event.start_time),
                             width     = event.bar_width,
                             height    = (event.start_time-event.end_time),
                             fillColor = getrlcolor(event.color[0],event.color[1],event.color[2])))
            label = Label()
            label.angle = event.label_angle
            label.setText(event.text)
            #label.fontName = 'Times-Roman'
            label.fontSize = 14
            if label.angle == 0:
                if event.label_placement == 'c':
                    label.setOrigin(event.x_position+(event.bar_width/2.0), yUp(self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0)))
                    label.boxAnchor = 'c'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    label.setOrigin(event.x_position-4, yUp(self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0)))
                    label.boxAnchor = 'e'
                    label.textAnchor = 'end'
                elif event.label_placement== 'r': 
                    label.setOrigin(event.x_position+event.bar_width+4, yUp(self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0)))
                    label.boxAnchor = 'w'
                    label.textAnchor = 'start'
                elif event.label_placement== 't': 
                    label.setOrigin(event.x_position+(event.bar_width/2.0), yUp(self.y_offset+event.end_time-4))
                    label.boxAnchor = 's'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'b': 
                    label.setOrigin(event.x_position+(event.bar_width/2.0), yUp(self.y_offset+event.start_time+4))
                    label.boxAnchor = 'n'
                    label.textAnchor = 'middle'
            else: # assume label_angle = 90
                if event.label_placement == 'c':
                    label.setOrigin(event.x_position+(event.bar_width/2.0), yUp(self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0)))
                    label.boxAnchor = 'c'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'l': 
                    label.setOrigin(event.x_position-4, yUp(self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0)))
                    label.boxAnchor = 's'
                    label.textAnchor = 'middle'
                elif event.label_placement== 'r': 
                    label.setOrigin(event.x_position+event.bar_width+4, yUp(self.y_offset+event.end_time+((event.start_time-event.end_time)/2.0)))
                    label.boxAnchor = 'n'
                    label.textAnchor = 'middle'
                elif event.label_placement== 't': 
                    label.setOrigin(event.x_position+(event.bar_width/2.0), yUp(self.y_offset+event.end_time-4))
                    label.boxAnchor = 'w'
                    label.textAnchor = 'start'
                elif event.label_placement== 'b': 
                    label.setOrigin(event.x_position+(event.bar_width/2.0), yUp(self.y_offset+event.start_time+4))
                    label.boxAnchor = 'e'
                    label.textAnchor = 'end'
            drawing.add(label)

    def getGeologicTime(self):
        geologic_time = TimeUnit(self,"(180,180,180)","Geologic Time",4567.2,0.7,0.0,0.0,"-","Base derived from moon rock and meteorite dates","-","from moon rock and meteorite dates","",8,-90,20.0,60.0,20)
        geologic_time.updatePositioningAndDimensions()
        
        phanerozoic_eon = TimeUnit(self,"(179,226,209)","Phanerozoic Eon",542.0,1.0,0.0,0.0,"U-Pb age from Oman coinciding with the negative carbon excursion","Trace fossil, lowest occurrence of 'Treptichnus (Phycodes) pedum'. Near base of negative carbon-isotope excursion","2.4 m above base of Member 2 of Chapel Island Fm., Fortune Head, Burin Peninsula, southeast Newfoundland, Canada","Ratified 1992","Episodes 17 (1 and 2), p.95-100, 1994.",14,-90,40.0,60.0,20)
        phanerozoic_eon.updatePositioningAndDimensions()
        geologic_time.add(phanerozoic_eon)
        
        cenozoic_era = TimeUnit(self,"(255,255,0)","Cenozoic Era",65.5,0.3,0.0,0.0,"Ar-Ar and U-Pb age agreement","Iridium geochemical anomaly. Associated with a major extinction horizon (foraminifers, calcareous nannofossils, dinosaurs, etc.);","Base of boundary clay, El Kef, Tunisia (but deterioration may require assigning a replacement section)","Ratified 1991","-",14,-90,60.0,60.0,20)
        cenozoic_era.updatePositioningAndDimensions()
        phanerozoic_eon.add(cenozoic_era)
        
        neogene_system = TimeUnit(self,"(253,204,138)","Neogene System",23.0,0.0,0.0,0.0,"Astronomical cycles in sediments","Base of magnetic polarity chronozone C6Cn.2n; lowest occurrence of planktonic foraminifer 'Paragloborotalia kugleri'; near extinction of calcareous nannofossil 'Reticulofenestra bisecta' (base Zone NN1).","35 m from top of Lemme-Carrosio section, Carrosio village, north of Genoa, Italy","Ratified 1996","Episodes 20 (1),p. 23-28, 1997",20,0,80.0,60.0,20)
        neogene_system.updatePositioningAndDimensions()
        cenozoic_era.add(neogene_system)
        
        holocene_series = TimeUnit(self,"(255,255,179)","Holocene Series",0.0,0.0,0.0,0.0,"Carbon-14 dating calibration","exactly 10,000 Carbon-14 years (= 11.5 ka calendar years BP) at the end of the Younger Dryas cold spell","-","Informal working definition","-",3,0,100.0,60.0,50)
        holocene_series.updatePositioningAndDimensions()
        neogene_system.add(holocene_series)
        
        pleistocene_series = TimeUnit(self,"(255,235,98)","Pleistocene Series",1.8,0.0,0.0,0.0,"Astronomical cycles in sediments","Just above top of magnetic polarity chronozone C2n (Olduvai) and the extinction level of calcareous nannofossil 'Discoaster brouweri' (base Zone CN13). Above are lowest occurrence of calcareous nannofossil medium 'Gephyrocapsa' spp. and extinction level of planktonic foraminifer 'Globigerinoides extremus'.","Top of sapropel layer 'e', Vrica section, Calabria, Italy","Ratified 1985","Episodes 8 (2), p.116-120, 1985",3,0,100.0,60.0,50)
        pleistocene_series.updatePositioningAndDimensions()
        neogene_system.add(pleistocene_series)
        
        upper_pleistocene_subseries = TimeUnit(self,"(255,235,98)","upper Pleistocene Subseries",0.1,0.0,0.0,0.0,"Astronomical cycles in sediments","base of the Eemian interglacial Stage (= base of marine isotope stage 5e) before final glacial episode of Pleistocene","Potentially, within sediment core under the Netherlands (Eemian type area)","Informal working definition","-",3,0,150.0,60.0,100)
        upper_pleistocene_subseries.updatePositioningAndDimensions()
        pleistocene_series.add(upper_pleistocene_subseries)
        
        middle_pleistocene_subseries = TimeUnit(self,"(255,235,98)","middle Pleistocene Subseries",0.8,0.0,0.1,0.0,"Astronomical cycles in sediments","Brunhes-Matuyama magnetic reversal","-","Informal working definition","-",3,0,150.0,60.1,100)
        middle_pleistocene_subseries.updatePositioningAndDimensions()
        pleistocene_series.add(middle_pleistocene_subseries)
        
        lower_pleistocene_subseries = TimeUnit(self,"(255,235,98)","lower Pleistocene Subseries",1.8,0.0,0.8,0.0,"Astronomical cycles in sediments","Just above top of magnetic polarity chronozone C2n (Olduvai) and the extinction level of calcareous nannofossil 'Discoaster brouweri' (base Zone CN13). Above are lowest occurrence of calcareous nannofossil medium 'Gephyrocapsa' spp. and extinction level of planktonic foraminifer 'Globigerinoides extremus'.","Top of sapropel layer 'e', Vrica section, Calabria, Italy","Ratified 1985","Episodes 8 (2), p.116-120, 1985",3,0,150.0,60.8,100)
        lower_pleistocene_subseries.updatePositioningAndDimensions()
        pleistocene_series.add(lower_pleistocene_subseries)
        
        pliocene_series = TimeUnit(self,"(254,235,172)","Pliocene Series",5.3,0.0,1.8,0.0,"Astronomical cycles in sediments","Top of magnetic polarity chronozone C3r, ~100 kyr before Thvera normal-polarity subchronozone (C3n.4n). Calcareous nannofossils -- near extinction level of 'Triquetrorhabdulus rugosus' (base Zone CN10b) and the lowest occurrence of 'Ceratolithus acutus'.","Base of Trubi Fm (base of carbonate cycle 1), Eraclea Minoa, Sicily, Italy","Ratified 2000","Episodes 23 (3), p.179-187, 2000",3,0,100.0,61.8,50)
        pliocene_series.updatePositioningAndDimensions()
        neogene_system.add(pliocene_series)
        
        gelasian_stage = TimeUnit(self,"(254,235,172)","Gelasian Stage",2.6,0.0,1.8,0.0,"Astronomical cycles in sediments","Isotopic stage 103, base of magnetic polarity chronozone C2r (Matuyama).  Above are extinction levels of calcareous nannofossil 'Discoaster pentaradiatus' and 'D. surculus' (base Zone CN12c).","Midpoint of sapropelic Nicola Bed ('A5'), Monte San Nicola, Gela, Sicily, Italy","Ratified 1996","Episodes 21 (2), p.82-87, 1998",3,0,150.0,61.8,100)
        gelasian_stage.updatePositioningAndDimensions()
        pliocene_series.add(gelasian_stage)
        
        piacenzian_stage = TimeUnit(self,"(254,235,172)","Piacenzian Stage",3.6,0.0,2.6,0.0,"Astronomical cycles in sediments","Base of magnetic polarity chronozone C2An (Gauss); extinction levels of planktonic foraminifers 'Globorotalia margaritae' (base Zone PL3) and 'Pulleniatina primalis'.","Base of beige layer of carbonate cycle 77, Punta Piccola, Sicily, Italy","Ratified 1997","Episodes 21 (2), p.88-93, 1998",3,0,150.0,62.6,100)
        piacenzian_stage.updatePositioningAndDimensions()
        pliocene_series.add(piacenzian_stage)
        
        zanclean_stage = TimeUnit(self,"(254,235,172)","Zanclean Stage",5.3,0.0,3.6,0.0,"Astronomical cycles in sediments","Top of magnetic polarity chronozone C3r, ~100 kyr before Thvera normal-polarity subchronozone (C3n.4n). Calcareous nannofossils -- near extinction level of 'Triquetrorhabdulus rugosus' (base Zone CN10b) and the lowest occurrence of 'Ceratolithus acutus'.","Base of Trubi Fm (base of carbonate cycle 1), Eraclea Minoa, Sicily, Italy","Ratified 2000","Episodes 23 (3), p.179-187, 2000",3,0,150.0,63.6,100)
        zanclean_stage.updatePositioningAndDimensions()
        pliocene_series.add(zanclean_stage)
        
        miocene_series = TimeUnit(self,"(255,222,0)","Miocene Series",23.0,0.0,5.3,0.0,"Astronomical cycles in sediments","Base of magnetic polarity chronozone C6Cn.2n; lowest occurrence of planktonic foraminifer 'Paragloborotalia kugleri'; near extinction of calcareous nannofossil 'Reticulofenestra bisecta' (base Zone NN1).","35 m from top of Lemme-Carrosio section, Carrosio village, north of Genoa, Italy","Ratified 1996","Episodes 20 (1),p. 23-28, 1997",3,0,100.0,65.3,50)
        miocene_series.updatePositioningAndDimensions()
        neogene_system.add(miocene_series)
        
        messinian_stage = TimeUnit(self,"(255,222,0)","Messinian Stage",7.2,0.0,5.3,0.0,"Astronomical cycles in sediments","Astrochronology age of 7.251 Ma; middle of magnetic polarity chronozone C3Br.1r; lowest regular occurrence of the 'Globorotalia conomiozea' planktonic foraminifer group.","Base of red layer of carbonate cycle 15, Oued Akrech, Rabat, Morocco","Ratified 2000","Episodes 23 (3), p. 172-178, 2000",3,0,150.0,65.3,100)
        messinian_stage.updatePositioningAndDimensions()
        miocene_series.add(messinian_stage)
        
        tortonian_stage = TimeUnit(self,"(255,222,0)","Tortonian Stage",11.6,0.0,7.2,0.0,"Astronomical cycles in sediments","Last Common Occurrences of the calcareous nannofossil 'Discoaster kugleri' and the planktonic foraminifer 'Globigerinoides subquadratus'. Associated with the short normal-polarity subchron C5r.2n.","Midpoint of sapropel 76, Monte dei Corvi beach section, Ancona, Italy","Ratified 2003","Episodes article in preparation",3,0,150.0,67.2,100)
        tortonian_stage.updatePositioningAndDimensions()
        miocene_series.add(tortonian_stage)
        
        serravallian_stage = TimeUnit(self,"(255,222,0)","Serravallian Stage",13.7,0.0,11.6,0.0,"Astronomical cycles in sediments","Near lowest occurrence of nannofossil 'Sphenolithus heteromorphus', and within magnetic polarity chronozone C5ABr.","-","GSSP anticipated in 2004","-",3,0,150.0,71.6,100)
        serravallian_stage.updatePositioningAndDimensions()
        miocene_series.add(serravallian_stage)
        
        langhian_stage = TimeUnit(self,"(255,222,0)","Langhian Stage",16.0,0.0,13.7,0.0,"Calibrated magnetic anomaly scale","Near first occurrence of planktonic foraminifer 'Praeorbulina glomerosa' and top of magnetic polarity chronozone C5Cn.1n","-","GSSP anticipated in 2004","-",3,0,150.0,73.7,100)
        langhian_stage.updatePositioningAndDimensions()
        miocene_series.add(langhian_stage)
        
        burdigalian_stage = TimeUnit(self,"(255,222,0)","Burdigalian Stage",20.4,0.0,16.0,0.0,"Calibrated magnetic anomaly scale","Near lowest occurrence of planktonic foraminifer 'Globigerinoides altiaperturus' or near top of magnetic polarity chronozone C6An","-","Guide event is undecided","-",3,0,150.0,76.0,100)
        burdigalian_stage.updatePositioningAndDimensions()
        miocene_series.add(burdigalian_stage)
        
        aquitanian_stage = TimeUnit(self,"(255,222,0)","Aquitanian Stage",23.0,0.0,20.4,0.0,"Astronomical cycles in sediments","Base of magnetic polarity chronozone C6Cn.2n; lowest occurrence of planktonic foraminifer 'Paragloborotalia kugleri'; near extinction of calcareous nannofossil 'Reticulofenestra bisecta' (base Zone NN1).","35 m from top of Lemme-Carrosio section, Carrosio village, north of Genoa, Italy","Ratified 1996","Episodes 20 (1),p. 23-28, 1997",3,0,150.0,80.4,100)
        aquitanian_stage.updatePositioningAndDimensions()
        miocene_series.add(aquitanian_stage)
        
        paleogene_system = TimeUnit(self,"(255,179,0)","Paleogene System",65.5,0.3,23.0,0.0,"Ar-Ar and U-Pb age agreement","Iridium geochemical anomaly. Associated with a major extinction horizon (foraminifers, calcareous nannofossils, dinosaurs, etc.);","Base of boundary clay, El Kef, Tunisia (but deterioration may require assigning a replacement section)","Ratified 1991","-",20,0,80.0,83.0,20)
        paleogene_system.updatePositioningAndDimensions()
        cenozoic_era.add(paleogene_system)
        
        oligocene_series = TimeUnit(self,"(234,198,114)","Oligocene Series",33.9,0.1,23.0,0.0,"Calibrated magnetic anomaly scale relative to base-Miocene and C24n.","Planktonic foraminifer, extinction of 'Hantkenina'","Base of marl bed at 19m above base of Massignano quarry, Ancona, Italy","Ratified 1992","Episodes 16 (3), p.379-382, 1993",3,0,100.0,83.0,50)
        oligocene_series.updatePositioningAndDimensions()
        paleogene_system.add(oligocene_series)
        
        chattian_stage = TimeUnit(self,"(234,198,114)","Chattian Stage",28.4,0.1,23.0,0.0,"Calibrated magnetic anomaly scale relative to base-Miocene and C24n. Arbitrary 100 kyr uncertainty assigned.","Planktonic foraminifer, extinction of 'Chiloguembelina' (base Zone P21b)","Probably in Umbria-Marche region of Italy","GSSP anticipated in 2004","-",3,0,150.0,83.0,100)
        chattian_stage.updatePositioningAndDimensions()
        oligocene_series.add(chattian_stage)
        
        rupelian_stage = TimeUnit(self,"(234,198,114)","Rupelian Stage",33.9,0.1,28.4,0.1,"Calibrated magnetic anomaly scale relative to base-Miocene and C24n.","Planktonic foraminifer, extinction of 'Hantkenina'","Base of marl bed at 19m above base of Massignano quarry, Ancona, Italy","Ratified 1992","Episodes 16 (3), p.379-382, 1993",3,0,150.0,88.4,100)
        rupelian_stage.updatePositioningAndDimensions()
        oligocene_series.add(rupelian_stage)
        
        eocene_series = TimeUnit(self,"(234,173,67)","Eocene Series",55.8,0.2,33.9,0.1,"Astronomical cycles in sediments scaled from base-Paleocene","Base of negative carbon-isotope excursion","Dababiya Section near Luxor, Egypt","Ratified 2003","Micropaleontology v.49 (Suppl. 1), 2003. Episodes article in preparation",3,0,100.0,93.9,50)
        eocene_series.updatePositioningAndDimensions()
        paleogene_system.add(eocene_series)
        
        priabonian_stage = TimeUnit(self,"(234,173,67)","Priabonian Stage",37.2,0.1,33.9,0.1,"Calibrated magnetic anomaly scale relative to base-Miocene and C24n","Near lowest occurrence of calcareous nannofossil 'Chiasmolithus oamaruensis' (base Zone NP18)","Probably in Umbria-Marche region of Italy","-","-",3,0,150.0,93.9,100)
        priabonian_stage.updatePositioningAndDimensions()
        eocene_series.add(priabonian_stage)
        
        bartonian_stage = TimeUnit(self,"(234,173,67)","Bartonian Stage",40.4,0.2,37.2,0.1,"Calibrated magnetic anomaly scale relative to base-Miocene and C24n","Near extinction of calcareous nannofossil 'Reticulofenestra reticulata'","-","-","-",3,0,150.0,97.2,100)
        bartonian_stage.updatePositioningAndDimensions()
        eocene_series.add(bartonian_stage)
        
        lutetian_stage = TimeUnit(self,"(234,173,67)","Lutetian Stage",48.6,0.2,40.4,0.2,"Calibrated magnetic anomaly scale relative to base-Miocene and C24n","Planktonic foraminifer, lowest occurrence of 'Hantkenina'","Leading candidate is Fortuna section, Murcia province, Betic Cordilleras, Spain","GSSP anticipated in 2004","-",3,0,150.0,100.4,100)
        lutetian_stage.updatePositioningAndDimensions()
        eocene_series.add(lutetian_stage)
        
        ypresian_stage = TimeUnit(self,"(234,173,67)","Ypresian Stage",55.8,0.2,48.6,0.2,"Astronomical cycles in sediments scaled from base-Paleocene","Base of negative carbon-isotope excursion","Dababiya Section near Luxor, Egypt","Ratified 2003","Micropaleontology v.49 (Suppl. 1), 2003. Episodes article in preparation",3,0,150.0,108.6,100)
        ypresian_stage.updatePositioningAndDimensions()
        eocene_series.add(ypresian_stage)
        
        paleocene_series = TimeUnit(self,"(235,147,1)","Paleocene Series",65.5,0.3,55.8,0.2,"Ar-Ar and U-Pb age agreement","Iridium geochemical anomaly. Associated with a major extinction horizon (foraminifers, calcareous nannofossils, dinosaurs, etc.);","Base of boundary clay, El Kef, Tunisia (but deterioration may require assigning a replacement section)","Ratified 1991","-",3,0,100.0,115.8,50)
        paleocene_series.updatePositioningAndDimensions()
        paleogene_system.add(paleocene_series)
        
        thanetian_stage = TimeUnit(self,"(235,147,1)","Thanetian Stage",58.7,0.2,55.8,0.2,"Astronomical cycles in sediments scaled from base Paleocene, using base of magnetic polarity chronozone C26n. Arbitrary 0.1 (2 precession cycles, plus the base-Paleogene radiometric) uncertainty assigned to all estimates.","Magnetic polarity chronozone, base of C26n, is a temporary assignment","Leading candidate is Zumaya section, northern Spain","Guide event is undecided","-",3,0,150.0,115.8,100)
        thanetian_stage.updatePositioningAndDimensions()
        paleocene_series.add(thanetian_stage)
        
        selandian_stage = TimeUnit(self,"(235,147,1)","Selandian Stage",61.7,0.2,58.7,0.2,"Astronomical cycles in sediments scaled from base Paleocene, using magnetic polarity chronozone placement of C27n.9","Boundary task group is considering a higher level -- base of calcareous nannofossil zone NP5 -- which would be ~1 myr younger.","Leading candidate is Zumaya section, northern Spain","Guide event is undecided","-",3,0,150.0,118.7,100)
        selandian_stage.updatePositioningAndDimensions()
        paleocene_series.add(selandian_stage)
        
        danian_stage = TimeUnit(self,"(235,147,1)","Danian Stage",65.5,0.3,61.7,0.2,"Ar-Ar and U-Pb age agreement","Iridium geochemical anomaly. Associated with a major extinction horizon (foraminifers, calcareous nannofossils, dinosaurs, etc.);","Base of boundary clay, El Kef, Tunisia (but deterioration may require assigning a replacement section)","Ratified 1991","-",3,0,150.0,121.7,100)
        danian_stage.updatePositioningAndDimensions()
        paleocene_series.add(danian_stage)
        
        mesozoic_era = TimeUnit(self,"(127,173,81)","Mesozoic Era",251.0,0.4,65.5,0.3,"Average of U-Pb constraints from Bowring et al. (1998)","Conodont, lowest occurrence of 'Hindeodus parvus'; termination of major negative carbon-isotope excursion. About 1 myr after peak of Late Permian extinctions.","Base of Bed 27c, Meishan, Zhejiang, China","Ratified 2001","Episodes 24 (2), p.102-114, 2001",14,-90,60.0,125.5,20)
        mesozoic_era.updatePositioningAndDimensions()
        phanerozoic_eon.add(mesozoic_era)
        
        cretaceous_system = TimeUnit(self,"(127,195,28)","Cretaceous System",145.5,4.0,65.5,0.3,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base of 'Berriasella jacobi' zone (M19n.2n.55)","Maybe near lowest occurrence of ammonite 'Berriasella jacobi'","-","Guide event is undecided","-",20,0,80.0,125.5,20)
        cretaceous_system.updatePositioningAndDimensions()
        mesozoic_era.add(cretaceous_system)
        
        upper_cretaceous_series = TimeUnit(self,"(222,241,151)","Upper Cretaceous Series",99.6,0.9,65.5,0.3,"Spline fit of Ar-Ar ages and ammonite zones, plus monitor standard correction. Then cycle stratigraphy to place foraminifer datum relative to ammonite zonation.","Planktonic foraminifer, lowest occurrence of 'Rotalipora globotruncanoides'","36 m below top of Marnes Bleues Formation, Mont Risou, Rosans, Haute-Alpes, SE France","Ratified 2002","Episodes 27 (1), p.21-32, 2004.",3,0,100.0,125.5,50)
        upper_cretaceous_series.updatePositioningAndDimensions()
        cretaceous_system.add(upper_cretaceous_series)
        
        maastrichtian_stage = TimeUnit(self,"(222,241,151)","Maastrichtian Stage",70.6,0.6,65.5,0.3,"Estimated placement relative to Ar-Ar calibrated Sr-curve","Mean of 12 biostratigraphic criteria of equal importance. Closely above is lowest occurrence of ammonite 'Pachydiscus neubergicus'. Boreal proxy is lowest occurrence of belemnite 'Belemnella lanceolata'.","115.2 m level in Grande Carriere quarry, Tercis-les-Bains, Landes province, SW France","Ratified 2001","Episodes 24 (4), p.229-238, 2001; Odin (ed.) IUGS Spec. Publ. Series, v.36, Elsevier, 910pp.",3,0,150.0,125.5,100)
        maastrichtian_stage.updatePositioningAndDimensions()
        upper_cretaceous_series.add(maastrichtian_stage)
        
        campanian_stage = TimeUnit(self,"(222,241,151)","Campanian Stage",83.5,0.7,70.6,0.6,"Spline fit of Ar-Ar ages and ammonite zones.","Crinoid, extinction of 'Marsupites testudinarius'","Leading candidates are in southern England and in Texas","-","-",3,0,150.0,130.6,100)
        campanian_stage.updatePositioningAndDimensions()
        upper_cretaceous_series.add(campanian_stage)
        
        santonian_stage = TimeUnit(self,"(222,241,151)","Santonian Stage",85.8,0.7,83.5,0.7,"Spline fit of Ar-Ar ages and ammonite zones.","Inoceramid bivalve, lowest occurrence of 'Cladoceramus undulatoplicatus'","Leading candidates are in Spain, England and Texas","-","-",3,0,150.0,143.5,100)
        santonian_stage.updatePositioningAndDimensions()
        upper_cretaceous_series.add(santonian_stage)
        
        coniacian_stage = TimeUnit(self,"(222,241,151)","Coniacian Stage",89.3,1.0,85.8,0.7,"Spline fit of Ar-Ar ages and ammonite zones.","Inoceramid bivalve, lowest occurrence of 'Cremnoceramus rotundatus' (sensu Troger non Fiege)","Base of Bed MK47, Salzgitter-Salder Quarry, SW of Hannover, Lower Saxony, northern Germany","GSSP anticipated in 2004","-",3,0,150.0,145.8,100)
        coniacian_stage.updatePositioningAndDimensions()
        upper_cretaceous_series.add(coniacian_stage)
        
        turonian_stage = TimeUnit(self,"(222,241,151)","Turonian Stage",93.5,0.8,89.3,1.0,"Spline fit of Ar-Ar ages and ammonite zones","Ammonite, lowest occurrence of 'Watinoceras devonense'","Base of Bed 120, Rock Canyon Anticline, east of Pueblo, Colorado, west-central USA","Ratified 2003","Episodes article in preparation",3,0,150.0,149.3,100)
        turonian_stage.updatePositioningAndDimensions()
        upper_cretaceous_series.add(turonian_stage)
        
        cenomanian_stage = TimeUnit(self,"(222,241,151)","Cenomanian Stage",99.6,0.9,93.5,0.8,"Spline fit of Ar-Ar ages and ammonite zones, plus monitor standard correction. Then cycle stratigraphy to place foraminifer datum relative to ammonite zonation.","Planktonic foraminifer, lowest occurrence of 'Rotalipora globotruncanoides'","36 m below top of Marnes Bleues Formation, Mont Risou, Rosans, Haute-Alpes, SE France","Ratified 2002","Episodes 27 (1), p.21-32, 2004.",3,0,150.0,153.5,100)
        cenomanian_stage.updatePositioningAndDimensions()
        upper_cretaceous_series.add(cenomanian_stage)
        
        lower_cretaceous_series = TimeUnit(self,"(179,223,127)","Lower Cretaceous Series",145.5,4.0,99.6,0.9,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base of 'Berriasella jacobi' zone (M19n.2n.55)","Maybe near lowest occurrence of ammonite 'Berriasella jacobi'","-","Guide event is undecided","-",3,0,100.0,159.6,50)
        lower_cretaceous_series.updatePositioningAndDimensions()
        cretaceous_system.add(lower_cretaceous_series)
        
        albian_stage = TimeUnit(self,"(179,223,127)","Albian Stage",112.0,1.0,99.6,0.9,"Estimated placement relative to bases of Cenomanian and Aptian, with large uncertainty due to lack of GSSP criteria. Ar-Ar age of 114.6 +/- 0.7 Ma from 'Parahoplites nutfieldensis' below.","Calcareous nannofossil, lowest occurrence of 'Praediscosphaera columnata' (= 'P. cretacea' of some earlier studies), is one potential marker.","-","Guide event is undecided","-",3,0,150.0,159.6,100)
        albian_stage.updatePositioningAndDimensions()
        lower_cretaceous_series.add(albian_stage)
        
        aptian_stage = TimeUnit(self,"(179,223,127)","Aptian Stage",125.0,1.0,112.0,1.0,"Base of M0r, as recomputed from Ar-Ar age from MIT guyot","Magnetic polarity chronozone, base of M0r","Leading candidate is Gorgo a Cerbara, Piobbico, Umbria-Marche, central Italy","-","-",3,0,150.0,172.0,100)
        aptian_stage.updatePositioningAndDimensions()
        lower_cretaceous_series.add(aptian_stage)
        
        barremian_stage = TimeUnit(self,"(179,223,127)","Barremian Stage",130.0,1.5,125.0,1.0,"Pacific spreading model for magnetic anomaly ages (variable rate), using placement at M5n.8.","Ammonite, lowest occurrence of 'Spitidiscus hugii' - 'Spitidiscus vandeckii' group","Leading candidate is Rio Argos near Caravaca, Murcia province, Spain","-","-",3,0,150.0,185.0,100)
        barremian_stage.updatePositioningAndDimensions()
        lower_cretaceous_series.add(barremian_stage)
        
        hauterivian_stage = TimeUnit(self,"(179,223,127)","Hauterivian Stage",136.4,2.0,130.0,1.5,"Pacific spreading model for magnetic anomaly ages (variable rate), using placement at base M11n","Ammonite, lowest occurrence of genus 'Acanthodiscus' (especially 'A. radiatus')","Leading candidate is La Charce village, Drome province, southeast France","-","-",3,0,150.0,190.0,100)
        hauterivian_stage.updatePositioningAndDimensions()
        lower_cretaceous_series.add(hauterivian_stage)
        
        valanginian_stage = TimeUnit(self,"(179,223,127)","Valanginian Stage",140.2,3.0,136.4,2.0,"Pacific spreading model for magnetic anomaly ages (variable rate), using placement at M14r.3 (base 'T.  pertransiens').","Calpionellid, lowest occurrence of 'Calpionellites darderi' (base of Calpionellid Zone E); followed by the lowest occurrence of ammonite 'Thurmanniceras pertransiens'","Leading candidate is near Montbrun-les-Bains, Drome province, southeast France","-","-",3,0,150.0,196.4,100)
        valanginian_stage.updatePositioningAndDimensions()
        lower_cretaceous_series.add(valanginian_stage)
        
        berriasian_stage = TimeUnit(self,"(179,223,127)","Berriasian Stage",145.5,4.0,140.2,3.0,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base of 'Berriasella jacobi' zone (M19n.2n.55)","Maybe near lowest occurrence of ammonite 'Berriasella jacobi'","-","Guide event is undecided","-",3,0,150.0,200.2,100)
        berriasian_stage.updatePositioningAndDimensions()
        lower_cretaceous_series.add(berriasian_stage)
        
        jurassic_system = TimeUnit(self,"(77,180,126)","Jurassic System",199.6,0.6,145.5,4.0,"U-Pb age just below proposed GSSP for base-Jurassic","Near lowest occurrence of smooth 'Psiloceras planorbis' ammonite group","-","Guide event is undecided","-",20,0,80.0,205.5,20)
        jurassic_system.updatePositioningAndDimensions()
        mesozoic_era.add(jurassic_system)
        
        upper_jurassic_series = TimeUnit(self,"(204,235,197)","Upper Jurassic Series",161.2,4.0,145.5,4.0,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base M36An","Ammonite, 'Brightia thuouxensis' Horizon at base of the 'Cardioceras scarburgense' Subzone (Quenstedtoceras mariae Zone)","Leading candidates are in SE France and southern England","GSSP anticipated in 2004","-",3,0,100.0,205.5,50)
        upper_jurassic_series.updatePositioningAndDimensions()
        jurassic_system.add(upper_jurassic_series)
        
        tithonian_stage = TimeUnit(self,"(204,235,197)","Tithonian Stage",150.8,4.0,145.5,4.0,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base M22An","Near base of 'Hybonoticeras hybonotum' ammonite zone and lowest occurrence of 'Gravesia' genus, and the base of magnetic polarity chronozone M22An","-","Guide event is undecided","-",3,0,150.0,205.5,100)
        tithonian_stage.updatePositioningAndDimensions()
        upper_jurassic_series.add(tithonian_stage)
        
        kimmeridgian_stage = TimeUnit(self,"(204,235,197)","Kimmeridgian Stage",155.7,4.0,150.8,4.0,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base M26r.2 (Boreal ammonite definition)","Ammonite, near base of 'Pictonia baylei' ammonite zone of Boreal realm","Leading candidates are In Scotland, SE France and Poland","GSSP anticipated in 2004","-",3,0,150.0,210.8,100)
        kimmeridgian_stage.updatePositioningAndDimensions()
        upper_jurassic_series.add(kimmeridgian_stage)
        
        oxfordian_stage = TimeUnit(self,"(204,235,197)","Oxfordian Stage",161.2,4.0,155.7,4.0,"Pacific spreading model for magnetic anomaly ages (variable rate), assigning to base M36An","Ammonite, 'Brightia thuouxensis' Horizon at base of the 'Cardioceras scarburgense' Subzone (Quenstedtoceras mariae Zone)","Leading candidates are in SE France and southern England","GSSP anticipated in 2004","-",3,0,150.0,215.7,100)
        oxfordian_stage.updatePositioningAndDimensions()
        upper_jurassic_series.add(oxfordian_stage)
        
        middle_jurassic_series = TimeUnit(self,"(127,202,147)","Middle Jurassic Series",175.6,2.0,161.2,4.0,"Duration of Aalenian-Toarcian from cycle stratigraphy","Ammonite, lowest occurrence of 'Leioceras' genus","base of Bed FZ107, Fuentelsalz, central Spain","Ratified 2000","Episodes 24 (3), p.166-175, 2001",3,0,100.0,221.2,50)
        middle_jurassic_series.updatePositioningAndDimensions()
        jurassic_system.add(middle_jurassic_series)
        
        callovian_stage = TimeUnit(self,"(127,202,147)","Callovian Stage",164.7,4.0,161.2,4.0,"Equal subzones scale Bajo-Bath-Callov","Ammonite, lowest occurrence of the genus 'Kepplerites' (Kosmoceratidae) (defines base of 'Macrocephalites herveyi' Zone in sub-Boreal province of Great Britain to southwest Germany)","Leading candidate is Pfeffingen, Swabian Alb, SW Germany","GSSP anticipated in 2004","-",3,0,150.0,221.2,100)
        callovian_stage.updatePositioningAndDimensions()
        middle_jurassic_series.add(callovian_stage)
        
        bathonian_stage = TimeUnit(self,"(127,202,147)","Bathonian Stage",167.7,3.5,164.7,4.0,"Equal subzones scale Bajo-Bath-Callov","Ammonite, lowest occurrence of 'Parkinsonia (G.) convergens' (defines base of 'Zigzagiceras zigzag' Zone)","-","-","-",3,0,150.0,224.7,100)
        bathonian_stage.updatePositioningAndDimensions()
        middle_jurassic_series.add(bathonian_stage)
        
        bajocian_stage = TimeUnit(self,"(127,202,147)","Bajocian Stage",171.6,3.0,167.7,3.5,"Equal subzones scale Bajo-Bath-Callov","Ammonite, lowest occurrence of the genus 'Hyperlioceras' (defines base of the 'Hyperlioceras discites' Zone)","Base of Bed AB11, 77.8 m above base of Murtinheira section, Cabo Mondego, western Portugal","Ratified 1996","Episodes 20 (1), p.16-22, 1997",3,0,150.0,227.7,100)
        bajocian_stage.updatePositioningAndDimensions()
        middle_jurassic_series.add(bajocian_stage)
        
        aalenian_stage = TimeUnit(self,"(127,202,147)","Aalenian Stage",175.6,2.0,171.6,3.0,"Duration of Aalenian-Toarcian from cycle stratigraphy","Ammonite, lowest occurrence of 'Leioceras' genus","base of Bed FZ107, Fuentelsalz, central Spain","Ratified 2000","Episodes 24 (3), p.166-175, 2001",3,0,150.0,231.6,100)
        aalenian_stage.updatePositioningAndDimensions()
        middle_jurassic_series.add(aalenian_stage)
        
        lower_jurassic_series = TimeUnit(self,"(102,192,146)","Lower Jurassic Series",199.6,0.6,175.6,2.0,"U-Pb age just below proposed GSSP for base-Jurassic","Near lowest occurrence of smooth 'Psiloceras planorbis' ammonite group","-","Guide event is undecided","-",3,0,100.0,235.6,50)
        lower_jurassic_series.updatePositioningAndDimensions()
        jurassic_system.add(lower_jurassic_series)
        
        toarcian_stage = TimeUnit(self,"(102,192,146)","Toarcian Stage",183.0,1.5,175.6,2.0,"Duration of Aalenian-Toarcian from cycle stratigraphy","Ammonite, near lowest occurrence of a diversified 'Eodactylites' ammonite fauna; correlates with the NW European Paltus horizon.","-","Guide event undecided","-",3,0,150.0,235.6,100)
        toarcian_stage.updatePositioningAndDimensions()
        lower_jurassic_series.add(toarcian_stage)
        
        pliensbachian_stage = TimeUnit(self,"(102,192,146)","Pliensbachian Stage",189.6,1.5,183.0,1.5,"Cycle-scaled linear Sr trend","Ammonite, lowest occurrences of 'Bifericeras donovani' and of genera 'Apoderoceras' and 'Gleviceras'.","Wine Haven section, Robin Hood's Bay, Yorkshire, England, UK","GSSP in ICS voting, July 2004","-",3,0,150.0,243.0,100)
        pliensbachian_stage.updatePositioningAndDimensions()
        lower_jurassic_series.add(pliensbachian_stage)
        
        sinemurian_stage = TimeUnit(self,"(102,192,146)","Sinemurian Stage",196.5,1.0,189.6,1.5,"Cycle-scaled linear Sr trend","Ammonite, lowest occurrence of arietitid genera 'Vermiceras' and 'Metophioceras'","0.9 m above base of Bed 145, East Quantoxhead, Watchet, West Somerset, SW England, UK","Ratified 2000","Episodes 25 (1), p. 22-26, 2002",3,0,150.0,249.6,100)
        sinemurian_stage.updatePositioningAndDimensions()
        lower_jurassic_series.add(sinemurian_stage)
        
        hettangian_stage = TimeUnit(self,"(102,192,146)","Hettangian Stage",199.6,0.6,196.5,1.0,"U-Pb age just below proposed GSSP for base-Jurassic","Near lowest occurrence of smooth 'Psiloceras planorbis' ammonite group","-","Guide event is undecided","-",3,0,150.0,256.5,100)
        hettangian_stage.updatePositioningAndDimensions()
        lower_jurassic_series.add(hettangian_stage)
        
        triassic_system = TimeUnit(self,"(103,195,183)","Triassic System",251.0,0.4,199.6,0.6,"Average of U-Pb constraints from Bowring et al. (1998)","Conodont, lowest occurrence of 'Hindeodus parvus'; termination of major negative carbon-isotope excursion. About 1 myr after peak of Late Permian extinctions.","Base of Bed 27c, Meishan, Zhejiang, China","Ratified 2001","Episodes 24 (2), p.102-114, 2001",20,0,80.0,259.6,20)
        triassic_system.updatePositioningAndDimensions()
        mesozoic_era.add(triassic_system)
        
        upper_triassic_series = TimeUnit(self,"(204,236,225)","Upper Triassic Series",228.0,2.0,199.6,0.6,"Magnetostratigraphic correlation to cycle-scaled Newark magnetic polarity pattern","Near first occurrence of the ammonoids 'Daxatina' or 'Trachyceras', and of the conodont 'Metapolygnathus polygnathiformis'","Candidate section at Prati di Stuores, Dolomites, northern Italy. Important reference sections in Spiti (India) and New Pass, Nevada (USA).","Guide event is undecided","-",3,0,100.0,259.6,50)
        upper_triassic_series.updatePositioningAndDimensions()
        triassic_system.add(upper_triassic_series)
        
        rhaetian_stage = TimeUnit(self,"(204,236,225)","Rhaetian Stage",203.6,1.5,199.6,0.6,"Magnetostratigraphic correlation to cycle-scaled Newark magnetic polarity pattern","Near lowest occurrence of ammonite 'Cochlocera', conodonts 'Misikella' spp. and 'Epigondolella mosheri', and radiolarian 'Proparvicingula moniliformis'","Key sections in Austria, British Columbia (Canada), and Turkey","Guide event is undecided","-",3,0,150.0,259.6,100)
        rhaetian_stage.updatePositioningAndDimensions()
        upper_triassic_series.add(rhaetian_stage)
        
        norian_stage = TimeUnit(self,"(204,236,225)","Norian Stage",216.5,2.0,203.6,1.5,"Magnetostratigraphic correlation to cycle-scaled Newark magnetic polarity pattern","Base of 'Klamathites macrolobatus' or 'Stikinoceras kerri' ammonoid zones and the 'Metapolygnathus communisti' or 'M. primitius' conodont zones","Leading candidates are in British Columbia (Canada), Sicily (Italy), and possibly Slovakia, Turkey (Antalya Taurus) and Oman.","Guide event is undecided","-",3,0,150.0,263.6,100)
        norian_stage.updatePositioningAndDimensions()
        upper_triassic_series.add(norian_stage)
        
        carnian_stage = TimeUnit(self,"(204,236,225)","Carnian Stage",228.0,2.0,216.5,2.0,"Magnetostratigraphic correlation to cycle-scaled Newark magnetic polarity pattern","Near first occurrence of the ammonoids 'Daxatina' or 'Trachyceras', and of the conodont 'Metapolygnathus polygnathiformis'","Candidate section at Prati di Stuores, Dolomites, northern Italy. Important reference sections in Spiti (India) and New Pass, Nevada (USA).","Guide event is undecided","-",3,0,150.0,276.5,100)
        carnian_stage.updatePositioningAndDimensions()
        upper_triassic_series.add(carnian_stage)
        
        middle_triassic_series = TimeUnit(self,"(153,215,190)","Middle Triassic Series",245.0,1.5,228.0,2.0,"Proportional subzonal scaling","Ammonite, near lowest occurrences of genera 'Japonites', 'Paradanubites', and 'Paracrochordiceras'; and of the conodont 'Chiosella timorensis'","Candidate section probable at Desli Caira, Dobrogea, Romania; significant sections in Guizhou Province (China).","GSSP anticipated in 2004","-",3,0,100.0,288.0,50)
        middle_triassic_series.updatePositioningAndDimensions()
        triassic_system.add(middle_triassic_series)
        
        ladinian_stage = TimeUnit(self,"(153,215,190)","Ladinian Stage",237.0,2.0,228.0,2.0,"U-Pb array by Mundil et al. on levels near 'Nevadites' (= 'Secedensis') ammonite zone in Dolomites, plus placement relative to magnetostratigraphy corrlations to cycle-scaled Newark magnetic polarity pattern","Alternate levels are near base of Reitzi, 'Secedensis', or 'Curionii' ammonite zone; near first occurrence of the conodont genus 'Budurovignathus'","Leading candidates are Bagolino (Italy) and Felsoons (Hungary). Important reference sections in the Humboldt Range, Nevada (USA).","Guide event is undecided","-",3,0,150.0,288.0,100)
        ladinian_stage.updatePositioningAndDimensions()
        middle_triassic_series.add(ladinian_stage)
        
        anisian_stage = TimeUnit(self,"(153,215,190)","Anisian Stage",245.0,1.5,237.0,2.0,"Proportional subzonal scaling","Ammonite, near lowest occurrences of genera 'Japonites', 'Paradanubites', and 'Paracrochordiceras'; and of the conodont 'Chiosella timorensis'","Candidate section probable at Desli Caira, Dobrogea, Romania; significant sections in Guizhou Province (China).","GSSP anticipated in 2004","-",3,0,150.0,297.0,100)
        anisian_stage.updatePositioningAndDimensions()
        middle_triassic_series.add(anisian_stage)
        
        lower_triassic_series = TimeUnit(self,"(103,179,159)","Lower Triassic Series",251.0,0.4,245.0,1.5,"Average of U-Pb constraints from Bowring et al. (1998)","Conodont, lowest occurrence of 'Hindeodus parvus'; termination of major negative carbon-isotope excursion. About 1 myr after peak of Late Permian extinctions.","Base of Bed 27c, Meishan, Zhejiang, China","Ratified 2001","Episodes 24 (2), p.102-114, 2001",3,0,100.0,305.0,50)
        lower_triassic_series.updatePositioningAndDimensions()
        middle_triassic_series.add(lower_triassic_series)
        
        olenekian_stage = TimeUnit(self,"(103,179,159)","Olenekian Stage",249.7,0.7,245.0,1.5,"Composite standard from conodonts scaled to base-Anisian and base-Triassic","Near lowest occurrence of 'Hedenstroemia' or 'Meekoceras gracilitatis' ammonites, and of the conodont 'Neospathodus waageni'","Candidate sections in Siberia (Russia) and probably Chaohu, Anhui Province, China. Important sections also in Spiti.","Guide event is undecided","-",3,0,150.0,305.0,100)
        olenekian_stage.updatePositioningAndDimensions()
        middle_triassic_series.add(olenekian_stage)
        
        induan_stage = TimeUnit(self,"(103,179,159)","Induan Stage",251.0,0.4,249.7,0.7,"Average of U-Pb constraints from Bowring et al. (1998)","Conodont, lowest occurrence of 'Hindeodus parvus'; termination of major negative carbon-isotope excursion. About 1 myr after peak of Late Permian extinctions.","Base of Bed 27c, Meishan, Zhejiang, China","Ratified 2001","Episodes 24 (2), p.102-114, 2001",3,0,150.0,309.7,100)
        induan_stage.updatePositioningAndDimensions()
        middle_triassic_series.add(induan_stage)
        
        paleozoic_era = TimeUnit(self,"(128,181,213)","Paleozoic Era",542.0,1.0,251.0,0.4,"U-Pb age from Oman coinciding with the negative carbon excursion","Trace fossil, lowest occurrence of 'Treptichnus (Phycodes) pedum'. Near base of negative carbon-isotope excursion","2.4 m above base of Member 2 of Chapel Island Fm., Fortune Head, Burin Peninsula, southeast Newfoundland, Canada","Ratified 1992","Episodes 17 (1 and 2), p.95-100, 1994.",14,-90,60.0,311.0,20)
        paleozoic_era.updatePositioningAndDimensions()
        phanerozoic_eon.add(paleozoic_era)
        
        permian_system = TimeUnit(self,"(103,198,221)","Permian System",299.0,0.8,251.0,0.4,"''","Conodont, lowest occurrence of 'Streptognathodus isolatus' within the 'S. wabaunsensis' conodont chronocline. 6 m higher is lowest fusilinid foraminifer 'Sphaeroschwagerina'","27 m above base of Bed 19, Aidaralash Creek, Aktobe, southern Ural Mountains, northern Kazakhstan","Ratified, 1996","Episodes 21 (1), p.11-18, 1998",20,0,80.0,311.0,20)
        permian_system.updatePositioningAndDimensions()
        paleozoic_era.add(permian_system)
        
        lopingian_series = TimeUnit(self,"(179,227,238)","Lopingian Series",260.4,0.7,251.0,0.4,"''","Conodont, lowest occurrence of conodont 'Clarkina postbitteri postbitteri' (Mei and Wardlaw); and termination of a major global sea-level lowstand","Base of Bed 6K/115 in Penglaitan section, S. bank of Hongshui River, 20 km ESE of Laibin country town, Guangxi Province, south China (midway between cities of Guilin and Nanning). Nearby Tieqiao (Rail-bridge) section is a supplementary reference section.","Ratified 2004","-",3,0,100.0,311.0,50)
        lopingian_series.updatePositioningAndDimensions()
        permian_system.add(lopingian_series)
        
        changhsingian_stage = TimeUnit(self,"(179,227,238)","Changhsingian Stage",253.8,0.7,251.0,0.4,"Permian-Carboniferous time scale is derived from calibrating a master composite section to selected radiometric ages","Near lowest occurrence of conodont 'Clarkina wangi'","Leading candidates are in China","-","-",3,0,150.0,311.0,100)
        changhsingian_stage.updatePositioningAndDimensions()
        lopingian_series.add(changhsingian_stage)
        
        wuchiapingian_stage = TimeUnit(self,"(179,227,238)","Wuchiapingian Stage",260.4,0.7,253.8,0.7,"''","Conodont, lowest occurrence of conodont 'Clarkina postbitteri postbitteri' (Mei and Wardlaw); and termination of a major global sea-level lowstand","Base of Bed 6K/115 in Penglaitan section, S. bank of Hongshui River, 20 km ESE of Laibin country town, Guangxi Province, south China (midway between cities of Guilin and Nanning). Nearby Tieqiao (Rail-bridge) section is a supplementary reference section.","Ratified 2004","-",3,0,150.0,313.8,100)
        wuchiapingian_stage.updatePositioningAndDimensions()
        lopingian_series.add(wuchiapingian_stage)
        
        guadalupian_series = TimeUnit(self,"(153,216,216)","Guadalupian Series",270.6,0.7,260.4,0.7,"''","Conodont, lowest occurrence of 'Jinogondolella nanginkensis'","42.7 m above base of Cutoff Formation, Stratotype Canyon, southern Guadalupe Mountains, Texas, USA","Ratified 2001","Episodes article in preparation",3,0,100.0,320.4,50)
        guadalupian_series.updatePositioningAndDimensions()
        permian_system.add(guadalupian_series)
        
        capitanian_stage = TimeUnit(self,"(153,216,216)","Capitanian Stage",265.8,0.7,260.4,0.7,"''","Conodont, lowest occurrence of 'Jinogondolella postserrata'","4.5 m above base of Pinery Limestone Member, Nipple Hill, SE Guadalupe Mountains, Texas, USA","Ratified 2001","Episodes article in preparation",3,0,150.0,320.4,100)
        capitanian_stage.updatePositioningAndDimensions()
        guadalupian_series.add(capitanian_stage)
        
        wordian_stage = TimeUnit(self,"(153,216,216)","Wordian Stage",268.0,0.7,265.8,0.7,"''","Conodont, lowest occurrence of 'Jinogondolella aserrata'","7.6 m above base of Getaway Ledge outcrop, Guadalupe Pass, SE Guadalupe Mountains, Texas, USA","Ratified 2001","Episodes article in preparation",3,0,150.0,325.8,100)
        wordian_stage.updatePositioningAndDimensions()
        guadalupian_series.add(wordian_stage)
        
        roadian_stage = TimeUnit(self,"(153,216,216)","Roadian Stage",270.6,0.7,268.0,0.7,"''","Conodont, lowest occurrence of 'Jinogondolella nanginkensis'","42.7 m above base of Cutoff Formation, Stratotype Canyon, southern Guadalupe Mountains, Texas, USA","Ratified 2001","Episodes article in preparation",3,0,150.0,328.0,100)
        roadian_stage.updatePositioningAndDimensions()
        guadalupian_series.add(roadian_stage)
        
        cisuralian_series = TimeUnit(self,"(128,206,201)","Cisuralian Series",299.0,0.8,270.6,0.7,"''","Conodont, lowest occurrence of 'Streptognathodus isolatus' within the 'S. wabaunsensis' conodont chronocline. 6 m higher is lowest fusilinid foraminifer 'Sphaeroschwagerina'","27 m above base of Bed 19, Aidaralash Creek, Aktbe, southern Ural Mountains, northern Kazakhstan","Ratified, 1996","Episodes 21 (1), p.11-18, 1998",3,0,100.0,330.6,50)
        cisuralian_series.updatePositioningAndDimensions()
        permian_system.add(cisuralian_series)
        
        kungurian_stage = TimeUnit(self,"(128,206,201)","Kungurian Stage",275.6,0.7,270.6,0.7,"''","Near lowest occurrence of conodont 'Neostreptognathus pnevi - N. exculptus'","Leading candidates are in southern Ural Mtns.","-","-",3,0,150.0,330.6,100)
        kungurian_stage.updatePositioningAndDimensions()
        cisuralian_series.add(kungurian_stage)
        
        artinskian_stage = TimeUnit(self,"(128,206,201)","Artinskian Stage",284.4,0.7,275.6,0.7,"''","Conodont, lowest occurrence of conodont 'Sweetognathus whitei'","Leading candidates are in southern Ural Mtns.","-","-",3,0,150.0,335.6,100)
        artinskian_stage.updatePositioningAndDimensions()
        cisuralian_series.add(artinskian_stage)
        
        sakmarian_stage = TimeUnit(self,"(128,206,201)","Sakmarian Stage",294.6,0.8,284.4,0.7,"''","Conodont, near lowest occurrence of conodont 'Sweetognathus merrelli'","Leading candidate is at Kondurovsky, Orenburg Province, Russia.","-","-",3,0,150.0,344.4,100)
        sakmarian_stage.updatePositioningAndDimensions()
        cisuralian_series.add(sakmarian_stage)
        
        asselian_stage = TimeUnit(self,"(128,206,201)","Asselian Stage",299.0,0.8,294.6,0.8,"''","Conodont, lowest occurrence of 'Streptognathodus isolatus' within the 'S. wabaunsensis' conodont chronocline. 6 m higher is lowest fusilinid foraminifer 'Sphaeroschwagerina'","27 m above base of Bed 19, Aidaralash Creek, Aktbe, southern Ural Mountains, northern Kazakhstan","Ratified, 1996","Episodes 21 (1), p.11-18, 1998",3,0,150.0,354.6,100)
        asselian_stage.updatePositioningAndDimensions()
        cisuralian_series.add(asselian_stage)
        
        carboniferous_system = TimeUnit(self,"(153,189,218)","Carboniferous System",359.2,2.5,299.0,0.8,"''","Conodont, above lowest occurrence of 'Siphonodella sulcata'","Base of Bed 89, La Serre, Montagne Noir, Cabrires, southern France","Ratified 1990","Episodes 14 (4), p.331-336, 1991",20,0,80.0,359.0,20)
        carboniferous_system.updatePositioningAndDimensions()
        paleozoic_era.add(carboniferous_system)
        
        pennsylvanian_subsystem = TimeUnit(self,"(104,159,202)","Pennsylvanian Subsystem",318.1,1.3,299.0,0.8,"''","Conodont, lowest occurrence of 'Declinognathodus nodiliferus' s.l.","82.9 m above top of Battleship Wash Fm., Arrow Canyon, southern Nevada, USA","GSSP ratified 1996. Subsystem rank with Mississippian and Pennsylvanian names ratified 2000.","Episodes 22 (4), p.272-283, 1999",3,0,150.0,359.0,100)
        pennsylvanian_subsystem.updatePositioningAndDimensions()
        carboniferous_system.add(pennsylvanian_subsystem)
        
        upper_pennsylvanian_series = TimeUnit(self,"(104,159,202)","Upper Pennsylvanian Series",306.5,1.0,299.0,0.8,"''","Near base of 'Obsoletes obsoletes' and 'Protriticites pseudomontiparus' fusulinid zone, or lowest occurrence of 'Parashumardites' ammonoid","-","Guide event is undecided","-",3,0,100.0,359.0,50)
        upper_pennsylvanian_series.updatePositioningAndDimensions()
        pennsylvanian_subsystem.add(upper_pennsylvanian_series)
        
        gzhelian_stage = TimeUnit(self,"(104,159,202)","Gzhelian Stage",303.9,0.9,299.0,0.8,"''","Near lowest occurrences of the fusulinids 'Daixina', 'Jigulites' and 'Rugosofusulina', or lowest occurrence of conodont 'Idiognathodus simulator' (s.str.). Close to lowest occurrence of ammonoid 'Shumardites'.","-","Guide event is undecided","-",3,0,150.0,359.0,100)
        gzhelian_stage.updatePositioningAndDimensions()
        upper_pennsylvanian_series.add(gzhelian_stage)
        
        kasimovian_stage = TimeUnit(self,"(104,159,202)","Kasimovian Stage",306.5,1.0,303.9,0.9,"''","Near base of 'Obsoletes obsoletes' and 'Protriticites pseudomontiparus' fusulinid zone, or lowest occurrence of 'Parashumardites ammonoid'","-","Guide event is undecided","-",3,0,150.0,363.9,100)
        kasimovian_stage.updatePositioningAndDimensions()
        upper_pennsylvanian_series.add(kasimovian_stage)
        
        middle_pennsylvanian_series = TimeUnit(self,"(104,159,202)","Middle Pennsylvanian Series",311.7,1.1,306.5,1.0,"''","Near lowest occurrences of 'Declinognathodus donetzianus' and/or 'Idiognathoides postsulcatus' conodont species, and fusulinid species 'Aljutovella aljutovica'","-","Guide event is undecided","-",3,0,100.0,366.5,50)
        middle_pennsylvanian_series.updatePositioningAndDimensions()
        pennsylvanian_subsystem.add(middle_pennsylvanian_series)
        
        moscovian_stage = TimeUnit(self,"(104,159,202)","Moscovian Stage",311.7,1.1,306.5,1.0,"''","Near lowest occurrences of 'Declinognathodus donetzianus' and/or 'Idiognathoides postsulcatus' conodont species, and fusulinid species 'Aljutovella aljutovica'","-","Guide event is undecided","-",3,0,150.0,366.5,100)
        moscovian_stage.updatePositioningAndDimensions()
        middle_pennsylvanian_series.add(moscovian_stage)
        
        lower_pennsylvanian_series = TimeUnit(self,"(104,159,202)","Lower Pennsylvanian Series",318.1,1.3,311.7,1.1,"''","Conodont, lowest occurrence of 'Declinognathodus nodiliferus' s.l.","82.9 m above top of Battleship Wash Fm., Arrow Canyon, southern Nevada, USA","GSSP ratified 1996. Subsystem rank with Mississippian and Pennsylvanian names ratified 2000.","Episodes 22 (4), p.272-283, 1999",3,0,100.0,371.7,50)
        lower_pennsylvanian_series.updatePositioningAndDimensions()
        middle_pennsylvanian_series.add(lower_pennsylvanian_series)
        
        bashkirian_stage = TimeUnit(self,"(104,159,202)","Bashkirian Stage",318.1,1.3,311.7,1.1,"''","Conodont, lowest occurrence of 'Declinognathodus nodiliferus' s.l.","82.9 m above top of Battleship Wash Fm., Arrow Canyon, southern Nevada, USA","GSSP ratified 1996. Subsystem rank with Mississippian and Pennsylvanian names ratified 2000.","Episodes 22 (4), p.272-283, 1999",3,0,150.0,371.7,100)
        bashkirian_stage.updatePositioningAndDimensions()
        middle_pennsylvanian_series.add(bashkirian_stage)
        
        mississippian_subsystem = TimeUnit(self,"(128,145,173)","Mississippian Subsystem",359.2,2.5,318.1,1.3,"''","Conodont, above lowest occurrence of 'Siphonodella sulcata'","Base of Bed 89, La Serre, Montagne Noir, Cabrires, southern France","Ratified 1990","Episodes 14 (4), p.331-336, 1991",3,0,150.0,378.1,100)
        mississippian_subsystem.updatePositioningAndDimensions()
        carboniferous_system.add(mississippian_subsystem)
        
        upper_mississippian_series = TimeUnit(self,"(128,145,173)","Upper Mississippian Series",326.4,1.6,318.1,1.3,"''","Near lowest occurrence of conodont, 'Lochriea crusiformis'","-","Guide event is undecided","-",3,0,100.0,378.1,50)
        upper_mississippian_series.updatePositioningAndDimensions()
        mississippian_subsystem.add(upper_mississippian_series)
        
        serpukhovian_stage = TimeUnit(self,"(128,145,173)","Serpukhovian Stage",326.4,1.6,318.1,1.3,"''","Near lowest occurrence of conodont, 'Lochriea crusiformis'","-","Guide event is undecided","-",3,0,150.0,378.1,100)
        serpukhovian_stage.updatePositioningAndDimensions()
        upper_mississippian_series.add(serpukhovian_stage)
        
        middle_mississippian_series = TimeUnit(self,"(128,145,173)","Middle Mississippian Series",345.3,2.1,326.4,1.6,"''","Foraminifer, lineage 'Eoparastaffella simplex' morphotype 1/morphotype 2","Leading candidate is Pengchong, south China","-","-",3,0,100.0,386.4,50)
        middle_mississippian_series.updatePositioningAndDimensions()
        mississippian_subsystem.add(middle_mississippian_series)
        
        visean_stage = TimeUnit(self,"(128,145,173)","Visean Stage",345.3,2.1,326.4,1.6,"''","Foraminifer, lineage 'Eoparastaffella simplex' morphotype 1/morphotype 2","Leading candidate is Pengchong, south China","-","-",3,0,150.0,386.4,100)
        visean_stage.updatePositioningAndDimensions()
        middle_mississippian_series.add(visean_stage)
        
        lower_mississippian_series = TimeUnit(self,"(128,145,173)","Lower Mississippian Series",359.2,2.5,345.3,2.1,"''","Conodont, above lowest occurrence of 'Siphonodella sulcata'","Base of Bed 89, La Serre, Montagne Noir, Cabrires, southern France","Ratified 1990","Episodes 14 (4), p.331-336, 1991",3,0,100.0,405.3,50)
        lower_mississippian_series.updatePositioningAndDimensions()
        mississippian_subsystem.add(lower_mississippian_series)
        
        tournaisian_stage = TimeUnit(self,"(128,145,173)","Tournaisian Stage",359.2,2.5,345.3,2.1,"''","Conodont, above lowest occurrence of 'Siphonodella sulcata'","Base of Bed 89, La Serre, Montagne Noir, Cabrires, southern France","Ratified 1990","Episodes 14 (4), p.331-336, 1991",3,0,150.0,405.3,100)
        tournaisian_stage.updatePositioningAndDimensions()
        lower_mississippian_series.add(tournaisian_stage)
        
        devonian_system = TimeUnit(self,"(153,153,201)","Devonian System",416.0,2.8,359.2,2.5,"base-Devonian from scale in Cooper (GTS 2004), which is 1 myr younger than Tucker et al (1998) estimate.","Graptolite, lowest occurrence of 'Monograptus uniformis'","Within Bed 20, Klonk, Barrandian area, southwest of Prague, Czech Republic","Ratified 1972","Martinsson (ed.), The Silurian-Devonian Boundary, IUGS Series A, no.5, 349 pp., 1977",20,0,80.0,419.2,20)
        devonian_system.updatePositioningAndDimensions()
        paleozoic_era.add(devonian_system)
        
        upper_devonian_series = TimeUnit(self,"(203,189,220)","Upper Devonian Series",385.3,2.6,359.2,2.5,"''","Conodont, lowest occurrence of 'Ancyrodella rotundiloba' (defines base of Lower 'Polygnathus asymmetricus' conodont Zone)","Base of Bed 42a', Col du Puech de la Suque section, St. Nazaire-de-Ladarez, SE Montagne Noir, southern France","Ratified 1986","Episodes 10 (2), p.97-101, 1987",3,0,100.0,419.2,50)
        upper_devonian_series.updatePositioningAndDimensions()
        devonian_system.add(upper_devonian_series)
        
        famennian_stage = TimeUnit(self,"(203,189,220)","Famennian Stage",374.5,2.6,359.2,2.5,"Devonian time scale is a statistical fit of a zonal composite using zonal data by M. House and composite data by Tucker et al. (1989).","Just above major extinction horizon (Upper Kellwasser Event), including conodonts 'Ancyrodella' and 'Ozarkodina' and goniatites of 'Gephuroceratidae' and 'Beloceratidae'","base of Bed 32a, upper Coumiac quarry, Cessenon, Montagne Noir, southern France","Ratified 1993","Episodes 16 (4), p.433-441, 1993",3,0,150.0,419.2,100)
        famennian_stage.updatePositioningAndDimensions()
        upper_devonian_series.add(famennian_stage)
        
        frasnian_stage = TimeUnit(self,"(203,189,220)","Frasnian Stage",385.3,2.6,374.5,2.6,"''","Conodont, lowest occurrence of 'Ancyrodella rotundiloba' (defines base of Lower 'Polygnathus asymmetricus' conodont Zone)","Base of Bed 42a', Col du Puech de la Suque section, St. Nazaire-de-Ladarez, SE Montagne Noir, southern France","Ratified 1986","Episodes 10 (2), p.97-101, 1987",3,0,150.0,434.5,100)
        frasnian_stage.updatePositioningAndDimensions()
        upper_devonian_series.add(frasnian_stage)
        
        middle_devonian_series = TimeUnit(self,"(153,131,190)","Middle Devonian Series",397.5,2.7,385.3,2.6,"''","Conodont, lowest occurrence of 'Polygnathus costatus partitus'; major faunal turnover","Base unit WP30, trench at Wetteldorf Richtschnitt, Schnecken-Wetteldorf, Eifel Hills, western Germany","Ratified 1985","Episodes 8 (2), p.104-109, 1985",3,0,100.0,445.3,50)
        middle_devonian_series.updatePositioningAndDimensions()
        devonian_system.add(middle_devonian_series)
        
        givetian_stage = TimeUnit(self,"(153,131,190)","Givetian Stage",391.8,2.7,385.3,2.6,"''","Conodont, lowest occurrence of 'Polygnathus hemiansatus', near base of goniatite 'Maenioceras' Stufe","Base of Bed 123, Jebel Mech Irdane ridge, Tafilalt, Morocco","Ratified 1994","Episodes 18 (3), p.107-115, 1995",3,0,150.0,445.3,100)
        givetian_stage.updatePositioningAndDimensions()
        middle_devonian_series.add(givetian_stage)
        
        eifelian_stage = TimeUnit(self,"(153,131,190)","Eifelian Stage",397.5,2.7,391.8,2.7,"''","Conodont, lowest occurrence of 'Polygnathus costatus partitus'; major faunal turnover","Base unit WP30, trench at Wetteldorf Richtschnitt, Schnecken-Wetteldorf, Eifel Hills, western Germany","Ratified 1985","Episodes 8 (2), p.104-109, 1985",3,0,150.0,451.8,100)
        eifelian_stage.updatePositioningAndDimensions()
        middle_devonian_series.add(eifelian_stage)
        
        lower_devonian_series = TimeUnit(self,"(128,125,186)","Lower Devonian Series",416.0,2.8,397.5,2.7,"base-Devonian from scale in Cooper (GTS 2004), which is 1 myr younger than Tucker et al (1998) estimate.","Graptolite, lowest occurrence of 'Monograptus uniformis'","Within Bed 20, Klonk, Barrandian area, southwest of Prague, Czech Republic","Ratified 1972","Martinsson (ed.), The Silurian-Devonian Boundary, IUGS Series A, no.5, 349 pp., 1977",3,0,100.0,457.5,50)
        lower_devonian_series.updatePositioningAndDimensions()
        devonian_system.add(lower_devonian_series)
        
        emsian_stage = TimeUnit(self,"(128,125,186)","Emsian Stage",407.0,2.8,397.5,2.7,"''","Conodont, lowest occurrence of 'Polygnathus kitabicus' (= 'Po. dehiscens')","Base of Bed 9/5, Zinzil'ban Gorge, SE of Samarkand, Uzbekistan","Ratified 1995","Episodes 20 (4), p. 235-240, 1997",3,0,150.0,457.5,100)
        emsian_stage.updatePositioningAndDimensions()
        lower_devonian_series.add(emsian_stage)
        
        pragian_stage = TimeUnit(self,"(128,125,186)","Pragian Stage",411.2,2.8,407.0,2.8,"''","Conodont, lowest occurrence of 'Eognathodus sulcatus'","Base of Bed 12, Valk Chuchle quarry, southwest part of Prague city, Czech Republic","Ratified 1989","Episodes 12 (2), p.109-113, 1989",3,0,150.0,467.0,100)
        pragian_stage.updatePositioningAndDimensions()
        lower_devonian_series.add(pragian_stage)
        
        lochkovian_stage = TimeUnit(self,"(128,125,186)","Lochkovian Stage",416.0,2.8,411.2,2.8,"base-Devonian from scale in Cooper (GTS 2004), which is 1 myr younger than Tucker et al (1998) estimate.","Graptolite, lowest occurrence of 'Monograptus uniformis'","Within Bed 20, Klonk, Barrandian area, southwest of Prague, Czech Republic","Ratified 1972","Martinsson (ed.), The Silurian-Devonian Boundary, IUGS Series A, no.5, 349 pp., 1977",3,0,150.0,471.2,100)
        lochkovian_stage.updatePositioningAndDimensions()
        lower_devonian_series.add(lochkovian_stage)
        
        silurian_system = TimeUnit(self,"(177,114,182)","Silurian System",443.7,1.5,416.0,2.8,"''","Graptolites, lowest occurrences of 'Parakidograptus acuminatus' and 'Akidograptus ascensus'","1.6 m above base of Birkhill Shale Fm., Dob's Linn, Moffat, Scotland, UK","Ratified 1984","Episodes 8 (2), p.98-100, 1985",20,0,80.0,476.0,20)
        silurian_system.updatePositioningAndDimensions()
        paleozoic_era.add(silurian_system)
        
        pridoli_series = TimeUnit(self,"(233,199,226)","Pridoli Series",418.7,2.7,416.0,2.8,"Silurian and Ordovician time scales are from calibrating a CONOP composite graptolite zonation to selecte radiometric ages","Graptolite, lowest occurrence of 'Monograptus parultimus'","Within Bed 96, Pozry section near Reporje, Barrandian area, Prague, Czech Republic","Ratified 1984","Episodes 8 (2), p.101-103, 1985",3,0,100.0,476.0,50)
        pridoli_series.updatePositioningAndDimensions()
        silurian_system.add(pridoli_series)
        
        ludlow_series = TimeUnit(self,"(202,167,209)","Ludlow Series",422.9,2.5,418.7,2.7,"''","Imprecise. Just below base of local acritarch 'Leptobrachion longhopense' range zone. May be near base of 'Neodiversograptus nilssoni' graptolite zone","Base of lithological unit F, Pitch Coppice quarry, Ludlow, Shropshire, southwest England, UK","Ratified 1980","Lethaia 14, p.168, 1981; Episodes 5 (3), p.21-23, 1982",3,0,100.0,478.7,50)
        ludlow_series.updatePositioningAndDimensions()
        silurian_system.add(ludlow_series)
        
        ludfordian_stage = TimeUnit(self,"(202,167,209)","Ludfordian Stage",421.3,2.6,418.7,2.7,"''","Imprecise. May be near base of 'Saetograptus leintwardinensis' graptolite zone.","Base of lithological unit C, Sunnyhill Quarry, Ludlow, Shropshire, southwest England, UK","Ratified 1980","Lethaia 14, p.168, 1981; Episodes 5 (3), p.21-23, 1982",3,0,150.0,478.7,100)
        ludfordian_stage.updatePositioningAndDimensions()
        ludlow_series.add(ludfordian_stage)
        
        gorstian_stage = TimeUnit(self,"(202,167,209)","Gorstian Stage",422.9,2.5,421.3,2.6,"''","Imprecise. Just below base of local acritarch 'Leptobrachion longhopense' range zone. May be near base of 'Neodiversograptus nilssoni' graptolite zone","Base of lithological unit F, Pitch Coppice quarry, Ludlow, Shropshire, southwest England, UK","Ratified 1980","Lethaia 14, p.168, 1981; Episodes 5 (3), p.21-23, 1982",3,0,150.0,481.3,100)
        gorstian_stage.updatePositioningAndDimensions()
        ludlow_series.add(gorstian_stage)
        
        wenlock_series = TimeUnit(self,"(177,137,179)","Wenlock Series",428.2,2.3,422.9,2.5,"''","Imprecise. Between the base of acritarch biozone 5 and extinction of conodont 'Pterospathodus amorphognathoides'. May be near base of 'Cyrtograptus centrifugus' graptolite zone.","Base of lithological unit G, Hughley Brook, Apedale, Shropshire, southwest England, UK","Ratified 1980","Lethaia 14, p.168, 1981; Episodes 5 (3), p.21-23, 1982",3,0,100.0,482.9,50)
        wenlock_series.updatePositioningAndDimensions()
        ludlow_series.add(wenlock_series)
        
        homerian_stage = TimeUnit(self,"(177,137,179)","Homerian Stage",426.2,2.4,422.9,2.5,"''","Graptolite, lowest occurrence of 'Cyrtograptus lundgreni' (defines base of 'C. lundgreni' graptolite zone)","Graptolite biozone intersection in stream section in Whitwell Coppice, Homer, Shropshire, southwest England, UK","Ratified 1980","Lethaia 14, p.168, 1981; Episodes 5 (3), p.21-23, 1982",3,0,150.0,482.9,100)
        homerian_stage.updatePositioningAndDimensions()
        ludlow_series.add(homerian_stage)
        
        sheinwoodian_stage = TimeUnit(self,"(177,137,179)","Sheinwoodian Stage",428.2,2.3,426.2,2.4,"''","Imprecise. Between the base of acritarch biozone 5 and extinction of conodont 'Pterospathodus amorphognathoides'. May be near base of 'Cyrtograptus centrifugus' graptolite zone.","Base of lithological unit G, Hughley Brook, Apedale, Shropshire, southwest England, UK","Ratified 1980","Lethaia 14, p.168, 1981; Episodes 5 (3), p.21-23, 1982",3,0,150.0,486.2,100)
        sheinwoodian_stage.updatePositioningAndDimensions()
        ludlow_series.add(sheinwoodian_stage)
        
        llandovery_series = TimeUnit(self,"(152,88,168)","Llandovery Series",443.7,1.5,428.2,2.3,"''","Graptolites, lowest occurrences of 'Parakidograptus acuminatus' and 'Akidograptus ascensus'","1.6 m above base of Birkhill Shale Fm., Dob's Linn, Moffat, Scotland, UK","Ratified 1984","Episodes 8 (2), p.98-100, 1985",3,0,100.0,488.2,50)
        llandovery_series.updatePositioningAndDimensions()
        ludlow_series.add(llandovery_series)
        
        telychian_stage = TimeUnit(self,"(152,88,168)","Telychian Stage",436.0,1.9,428.2,2.3,"''","Brachiopods, just above extinction of 'Eocoelia intermedia' and below lowest succeeding species 'Eocoelia curtisi'. Near base of 'Monograptus turriculatus' graptolite zone.","Locality 162 in transect d, Cefn Cerig road, Llandovery area, south-central Wales, UK","Ratified 1984","Episodes 8 (2), p.101-103, 1985",3,0,150.0,488.2,100)
        telychian_stage.updatePositioningAndDimensions()
        ludlow_series.add(telychian_stage)
        
        aeronian_stage = TimeUnit(self,"(152,88,168)","Aeronian Stage",439.0,1.8,436.0,1.9,"''","Graptolite, lowest occurrence of 'Monograptus austerus sequens' (defines base of 'Monograptus triangulatus' graptolite zone)","Base of locality 72 in transect h, Trefawr forestry road, north of Cwm-coed-Aeron Farm, Llandovery area, south-central Wales, UK","Ratified 1984","Episodes 8 (2), p.101-103, 1985",3,0,150.0,496.0,100)
        aeronian_stage.updatePositioningAndDimensions()
        ludlow_series.add(aeronian_stage)
        
        rhuddanian_stage = TimeUnit(self,"(152,88,168)","Rhuddanian Stage",443.7,1.5,439.0,1.8,"''","Graptolites, lowest occurrences of 'Parakidograptus acuminatus' and 'Akidograptus ascensus'","1.6 m above base of Birkhill Shale Fm., Dob's Linn, Moffat, Scotland, UK","Ratified 1984","Episodes 8 (2), p.98-100, 1985",3,0,150.0,499.0,100)
        rhuddanian_stage.updatePositioningAndDimensions()
        ludlow_series.add(rhuddanian_stage)
        
        ordovician_system = TimeUnit(self,"(249,129,166)","Ordovician System",488.3,1.7,443.7,1.5,"''","Conodont, lowest occurrence of 'Iapetognathus fluctivagus';just above base of 'Cordylodus lindstromi' conodont Zone. Just below lowest occurrence of planktonic graptolites. Currently dated around 489 Ma","Within Bed 23 at the 101.8 m level, Green Point, western Newfoundland, Canada","Ratified 2000","Episodes 24 (1), p.19-28,",20,0,80.0,503.7,20)
        ordovician_system.updatePositioningAndDimensions()
        paleozoic_era.add(ordovician_system)
        
        upper_ordovician_series = TimeUnit(self,"(251,180,189)","Upper Ordovician Series",460.9,1.6,443.7,1.5,"''","Graptolite, lowest occurrence of 'Nemagraptus gracilis'","1.4 m below phosphorite in E14a outcrop, Fgelsng, Scane, southern Sweden","Ratified 2002","Episodes 23 (2), p.102-109, 2000 (proposal; formal GSSP publication in preparation).",3,0,100.0,503.7,50)
        upper_ordovician_series.updatePositioningAndDimensions()
        ordovician_system.add(upper_ordovician_series)
        
        hirnantian_stage = TimeUnit(self,"(251,180,189)","Hirnantian Stage",445.6,1.5,443.7,1.5,"''","Potentially at base of the 'Normalograptus extraordinarius - N. ojsuensis' graptolite biozone","Candidate section is Wangjiawan, China","-","-",3,0,150.0,503.7,100)
        hirnantian_stage.updatePositioningAndDimensions()
        upper_ordovician_series.add(hirnantian_stage)
        
        sixth_ordovician_stage = TimeUnit(self,"(251,180,189)","sixth Ordovician Stage",455.8,1.6,445.6,1.5,"''","Potentially near first appearance of the graptolite 'Diplacanthograptus caudatus'","Candidate sections ae Black Knob Ridge (Oklahoma, USA) and Hartfell Spa (S. Scotland, UK)","Guide event is undecided","-",3,0,150.0,505.6,100)
        sixth_ordovician_stage.updatePositioningAndDimensions()
        upper_ordovician_series.add(sixth_ordovician_stage)
        
        fifth_ordovician_stage = TimeUnit(self,"(251,180,189)","fifth Ordovician Stage",460.9,1.6,455.8,1.6,"''","Graptolite, lowest occurrence of 'Nemagraptus gracilis'","1.4 m below phosphorite in E14a outcrop, Fgelsng, Scane, southern Sweden","Ratified 2002","Episodes 23 (2), p.102-109, 2000 (proposal; formal GSSP publication in preparation).",3,0,150.0,515.8,100)
        fifth_ordovician_stage.updatePositioningAndDimensions()
        upper_ordovician_series.add(fifth_ordovician_stage)
        
        middle_ordovician_series = TimeUnit(self,"(250,154,177)","Middle Ordovician Series",471.8,1.6,460.9,1.6,"''","Conodont, potentially lowest occurrence of 'Protoprioniodus aranda' or of 'Baltoniodus triangularis'","Candidate sections at Niquivil (Argentina) and Huanghuachang (China)","-","-",3,0,100.0,520.9,50)
        middle_ordovician_series.updatePositioningAndDimensions()
        ordovician_system.add(middle_ordovician_series)
        
        darriwillian_stage = TimeUnit(self,"(250,154,177)","Darriwillian Stage",468.1,1.6,460.9,1.6,"''","Graptolite, lowest occurrence of 'Undulograptus austrodentatus'","Base of Bed AEP184, 22 m below top of Ningkuo Fm., Huangnitang, Changshan, Zhejiang province, southeast China","Ratified 1987","Episodes 20 (3), p.158-166, 1997",3,0,150.0,520.9,100)
        darriwillian_stage.updatePositioningAndDimensions()
        middle_ordovician_series.add(darriwillian_stage)
        
        third_ordovician_stage = TimeUnit(self,"(250,154,177)","third Ordovician Stage",471.8,1.6,468.1,1.6,"''","Conodont, potentially lowest occurrence of 'Protoprioniodus aranda' or of 'Baltoniodus triangularis'","Candidate sections at Niquivil (Argentina) and Huanghuachang (China)","-","-",3,0,150.0,528.1,100)
        third_ordovician_stage.updatePositioningAndDimensions()
        middle_ordovician_series.add(third_ordovician_stage)
        
        lower_ordovician_series = TimeUnit(self,"(230,125,164)","Lower Ordovician Series",488.3,1.7,471.8,1.6,"''","Conodont, lowest occurrence of 'Iapetognathus fluctivagus';just above base of 'Cordylodus lindstromi' conodont Zone. Just below lowest occurrence of planktonic graptolites. Currently dated around 489 Ma","Within Bed 23 at the 101.8 m level, Green Point, western Newfoundland, Canada","Ratified 2000","Episodes 24 (1), p.19-28,",3,0,100.0,531.8,50)
        lower_ordovician_series.updatePositioningAndDimensions()
        ordovician_system.add(lower_ordovician_series)
        
        second_ordovician_stage = TimeUnit(self,"(230,125,164)","second Ordovician Stage",478.6,1.7,471.8,1.6,"''","Graptolite, lowest occurrence of 'Tetragraptus approximatus'","Just above E bed, Diabasbrottet quarry, Vstergtland, southern Sweden","Ratified 2002","Episodes article in preparation",3,0,150.0,531.8,100)
        second_ordovician_stage.updatePositioningAndDimensions()
        lower_ordovician_series.add(second_ordovician_stage)
    
        tremadocian_stage = TimeUnit(self,"(230,125,164)","Tremadocian Stage",488.3,1.7,478.6,1.7,"''","Conodont, lowest occurrence of 'Iapetognathus fluctivagus';just above base of 'Cordylodus lindstromi' conodont Zone. Just below lowest occurrence of planktonic graptolites. Currently dated around 489 Ma","Within Bed 23 at the 101.8 m level, Green Point, western Newfoundland, Canada","Ratified 2000","Episodes 24 (1), p.19-28,",3,0,150.0,538.6,100)
        tremadocian_stage.updatePositioningAndDimensions()
        lower_ordovician_series.add(tremadocian_stage)
        
        cambrian_system = TimeUnit(self,"(251,128,95)","Cambrian System",542.0,1.0,488.3,1.7,"U-Pb age from Oman coinciding with the negative carbon excursion","Trace fossil, lowest occurrence of 'Treptichnus (Phycodes) pedum'. Near base of negative carbon-isotope excursion","2.4 m above base of Member 2 of Chapel Island Fm., Fortune Head, Burin Peninsula, southeast Newfoundland, Canada","Ratified 1992","Episodes 17 (1 and 2), p.95-100, 1994.",20,0,80.0,548.3,20)
        cambrian_system.updatePositioningAndDimensions()
        paleozoic_era.add(cambrian_system)
        
        upper_cambrian_furongian_series = TimeUnit(self,"(253,205,184)","Upper Cambrian (Furongian) Series",501.0,2.0,488.3,1.7,"Radiometric ages near primary marker level. Estimated age and uncertainty only.","Trilobite, lowest occurrence of agnostoid 'Glyptagnostus reticulatus'. Coincides with base of large positive carbon-isotope excursion","369.06 m above base of Huaqiao Fm, Paibi section, NW Hunan province, south China","Ratified 2003","Episodes article in preparation",3,0,100.0,548.3,50)
        upper_cambrian_furongian_series.updatePositioningAndDimensions()
        cambrian_system.add(upper_cambrian_furongian_series)
        
        paibian_stage = TimeUnit(self,"(253,205,184)","Paibian Stage",501.0,2.0,488.3,1.7,"Radiometric ages near primary marker level. Estimated age and uncertainty only.","Trilobite, lowest occurrence of agnostoid 'Glyptagnostus reticulatus'. Coincides with base of large positive carbon-isotope excursion","369.06 m above base of Huaqiao Fm, Paibi section, NW Hunan province, south China","Ratified 2003","Episodes article in preparation",3,0,150.0,548.3,100)
        paibian_stage.updatePositioningAndDimensions()
        upper_cambrian_furongian_series.add(paibian_stage)
        
        middle_cambrian_series = TimeUnit(self,"(232,174,151)","Middle Cambrian Series",513.0,2.0,501.0,2.0,"Radiometric ages near primary marker level. Estimated age and uncertainty only","Potential GSSP levels in Middle Cambrian are based mainly on trilobites","-","-","-",3,0,100.0,561.0,50)
        middle_cambrian_series.updatePositioningAndDimensions()
        cambrian_system.add(middle_cambrian_series)
        
        lower_cambrian_series = TimeUnit(self,"(231,124,114)","Lower Cambrian Series",542.0,1.0,513.0,2.0,"U-Pb age from Oman coinciding with the negative carbon excursion","Trace fossil, lowest occurrence of 'Treptichnus (Phycodes) pedum'. Near base of negative carbon-isotope excursion","2.4 m above base of Member 2 of Chapel Island Fm., Fortune Head, Burin Peninsula, southeast Newfoundland, Canada","Ratified 1992","Episodes 17 (1 and 2), p.95-100, 1994.",3,0,100.0,573.0,50)
        lower_cambrian_series.updatePositioningAndDimensions()
        cambrian_system.add(lower_cambrian_series)
        
        proterozoic_eon = TimeUnit(self,"(204,216,145)","Proterozoic Eon",2500.0,-1.0,542.0,1.0,"Defined chronometrically","Base = 2500 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",14,-90,40.0,602.0,20)
        proterozoic_eon.updatePositioningAndDimensions()
        geologic_time.add(proterozoic_eon)
        
        neoproterozoic_era = TimeUnit(self,"(202,165,149)","Neoproterozoic Era",1000.0,-1.0,542.0,1.0,"Defined chronometrically","Base = 1000 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",14,-90,60.0,602.0,20)
        neoproterozoic_era.updatePositioningAndDimensions()
        proterozoic_eon.add(neoproterozoic_era)
        
        ediacaran_system = TimeUnit(self,"(234,216,188)","Ediacaran System",630.0,-1.0,542.0,1.0,"Age suggested by Ediacaran Subcomm.; bracketed by radiometric ages of 600 and 635 Ma","Termination of Varanger (or Marinoan) glaciation.","'Base of the Marinoan cap carbonate (Nuccaleena Formation), immediately above the Elatina diamictite in the Enorama Creek section, Flinders Ranges, South Australia.'","'Neoproterozoic III' (ratified 1990 with base defined chronometrically at 650 Ma) was formally replaced by Ediacaran Period and its GSSP in Feb 2004.","Episodes article in preparation",20,0,80.0,602.0,20)
        ediacaran_system.updatePositioningAndDimensions()
        neoproterozoic_era.add(ediacaran_system)
        
        cryogenian_system = TimeUnit(self,"(220,171,170)","Cryogenian System",850.0,-1.0,630.0,-1.0,"Defined chronometrically","Base = 850 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,690.0,20)
        cryogenian_system.updatePositioningAndDimensions()
        neoproterozoic_era.add(cryogenian_system)
        
        tonian_system = TimeUnit(self,"(203,164,108)","Tonian System",1000.0,-1.0,850.0,-1.0,"Defined chronometrically","Base = 1000 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,910.0,20)
        tonian_system.updatePositioningAndDimensions()
        neoproterozoic_era.add(tonian_system)
        
        mesoproterozoic_era = TimeUnit(self,"(221,194,136)","Mesoproterozoic Era",1600.0,-1.0,1000.0,-1.0,"Defined chronometrically","Base = 1600 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",14,-90,60.0,1060.0,20)
        mesoproterozoic_era.updatePositioningAndDimensions()
        proterozoic_eon.add(mesoproterozoic_era)
        
        stenian_system = TimeUnit(self,"(221,194,136)","Stenian System",1200.0,-1.0,1000.0,-1.0,"Defined chronometrically","Base = 1200 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,1060.0,20)
        stenian_system.updatePositioningAndDimensions()
        mesoproterozoic_era.add(stenian_system)
        
        ectasian_system = TimeUnit(self,"(221,194,136)","Ectasian System",1400.0,-1.0,1200.0,-1.0,"Defined chronometrically","Base = 1400 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,1260.0,20)
        ectasian_system.updatePositioningAndDimensions()
        mesoproterozoic_era.add(ectasian_system)
        
        calymmian_system = TimeUnit(self,"(221,194,136)","Calymmian System",1600.0,-1.0,1400.0,-1.0,"Defined chronometrically","Base = 1600 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,1460.0,20)
        calymmian_system.updatePositioningAndDimensions()
        mesoproterozoic_era.add(calymmian_system)
        
        paleoproterozoic_era = TimeUnit(self,"(179,178,94)","Paleoproterozoic Era",2500.0,-1.0,1600.0,-1.0,"Defined chronometrically","Base = 2500 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",14,-90,60.0,1660.0,20)
        paleoproterozoic_era.updatePositioningAndDimensions()
        proterozoic_eon.add(paleoproterozoic_era)
        
        statherian_system = TimeUnit(self,"(179,178,94)","Statherian System",1800.0,-1.0,1600.0,-1.0,"Defined chronometrically","Base = 1800 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,1660.0,20)
        statherian_system.updatePositioningAndDimensions()
        paleoproterozoic_era.add(statherian_system)
        
        orosirian_system = TimeUnit(self,"(179,178,94)","Orosirian System",2050.0,-1.0,1800.0,-1.0,"Defined chronometrically","Base = 2050 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,1860.0,20)
        orosirian_system.updatePositioningAndDimensions()
        paleoproterozoic_era.add(orosirian_system)
        
        rhyacian_system = TimeUnit(self,"(179,178,94)","Rhyacian System",2300.0,-1.0,2050.0,-1.0,"Defined chronometrically","Base = 2300 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,2110.0,20)
        rhyacian_system.updatePositioningAndDimensions()
        paleoproterozoic_era.add(rhyacian_system)
        
        siderian_system = TimeUnit(self,"(179,178,94)","Siderian System",2500.0,-1.0,2300.0,-1.0,"Defined chronometrically","Base = 2500 Ma","-","Ratified 1990","Episodes 14 (2), p.139-140, 1991",20,0,80.0,2360.0,20)
        siderian_system.updatePositioningAndDimensions()
        paleoproterozoic_era.add(siderian_system)
        
        archean_eon = TimeUnit(self,"(153,173,172)","Archean Eon",3800.0,-1.0,2500.0,-1.0,"Base not defined formally","-","-","Subcomm. decision 1996, but not submitted to ICS","Informally in Episodes 15 (2), p.122-123, 1992",14,-90,40.0,2560.0,20)
        archean_eon.updatePositioningAndDimensions()
        geologic_time.add(archean_eon)
        
        neoarchean_era = TimeUnit(self,"(203,205,200)","Neoarchean Era",2800.0,-1.0,2500.0,-1.0,"Defined chronometrically","Base = 2800 Ma","-","Subcomm. decision 1996, but not submitted to ICS","Informally in Episodes 15 (2), p.122-123, 1992.",14,-90,60.0,2560.0,20)
        neoarchean_era.updatePositioningAndDimensions()
        archean_eon.add(neoarchean_era)
        
        mesoarchean_era = TimeUnit(self,"(221,194,136)","Mesoarchean Era",3200.0,-1.0,2800.0,-1.0,"Defined chronometrically","Base = 3200 Ma","-","Subcomm. decision 1996, but not submitted to ICS","Informally in Episodes 15 (2), p.122-123, 1992.",14,-90,60.0,2860.0,20)
        mesoarchean_era.updatePositioningAndDimensions()
        archean_eon.add(mesoarchean_era)
        
        paleoarchean_era = TimeUnit(self,"(153,151,145)","Paleoarchean Era",3600.0,-1.0,3200.0,-1.0,"Defined chronometrically","Base = 3600 Ma","-","Subcomm. decision 1996, but not submitted to ICS","Informally in Episodes 15 (2), p.122-123, 1992.",14,-90,60.0,3260.0,20)
        paleoarchean_era.updatePositioningAndDimensions()
        archean_eon.add(paleoarchean_era)
        
        eoarchean_era = TimeUnit(self,"(128,144,144)","Eoarchean Era",3800.0,-1.0,3600.0,-1.0,"Base not defined formally","-","-","Subcomm. decision 1996, but not submitted to ICS","Informally in Episodes 15 (2), p.122-123, 1992",14,-90,60.0,3660.0,20)
        eoarchean_era.updatePositioningAndDimensions()
        archean_eon.add(eoarchean_era)
        
        hadean_eon = TimeUnit(self,"(255,128,128)","Hadean Eon",4567.2,0.7,3800.0,-1.0,"-","Base derived from moon rock and meteorite dates","-","from moon rock and meteorite dates","",14,-90,40.0,3860.0,20)
        hadean_eon.updatePositioningAndDimensions()
        geologic_time.add(hadean_eon)
        
        return geologic_time

class PointEvent:
    def __init__(self,label_placement,label_angle,x_position,symbol_type,symbol_width,symbol_height,time,font_size,text,color):
        self.label_placement = label_placement.lower()
        self.label_angle = label_angle
        self.x_position = x_position
        self.symbol_type = symbol_type
        self.symbol_width = symbol_width
        self.symbol_height = symbol_height
        self.time = time
        self.font_size = font_size
        self.text = text
        self.color = color

    def imageMapTag(self, y_offset):
        if self.symbol_type == 'ellipse':
            x1 = self.x_position + (self.symbol_width / 2.0)
            y1 = y_offset + self.time 
            radius = self.symbol_height

            return "<area onmouseover=\"return escape(\'%s\')\" shape=\"circle\" coords=\"%d, %d, %d\">\n" % (self.text, x1, y1, radius)
        elif self.symbol_type == 'rect':
            x1 = self.x_position
            y1 = y_offset + self.time
            x2 = x1 + self.symbol_width
            y2 = y_offset + self.time + self.symbol_height

            return "<area onmouseover=\"return escape(\'%s\')\" shape=\"rect\" coords=\"%d, %d, %d, %d\">\n" % (self.text, x1, y1, x2, y2)

    def htmlForListing(self):
        var = []
        var += "<dl>\n"
        var += "    <dt>%2.1f Ma:</dt>\n" % self.time 
        var += "    <dd>%s</dd>\n" % (self.text)
        var += "</dl>\n" 

        return var

class RangeEvent:
    def __init__(self,label_placement,label_angle,x_position,bar_width,start_time,end_time,font_size,text,color):
        self.label_placement = label_placement.lower()
        self.label_angle = label_angle
        self.x_position = x_position
        self.bar_width = bar_width
        self.start_time = start_time
        self.end_time = end_time
        self.font_size = font_size
        self.text = text
        self.color = color

    def imageMapTag(self, y_offset):
        x1 = self.x_position
        y1 = y_offset + self.end_time
        x2 = x1 + self.bar_width
        y2 = y_offset + self.start_time

        return "<area onmouseover=\"return escape(\'%s\')\" shape=\"rect\" coords=\"%d, %d, %d, %d\">\n" % (self.text, x1, y1, x2, y2)

    def htmlForListing(self):
        var = []
        var += "<dl>\n"
        var += "    <dt>%2.1f Ma - %2.1f Ma:</dt>\n" % (self.end_time, self.start_time)
        var += "    <dd>%s</dd>\n" % (self.text)
        var += "</dl>\n" 

        return var

class TimeUnit:
    def __init__(self,timescale,colorstring,
                 name,begin,begin_error,end,end_error,
                 age_derivation,events,gssp,
                 status,publications,
                 textsize,textrotate,
                 x,y,width):
        self.timescale = timescale
        self.name = name
        self.begin = begin
        self.begin_error = begin_error
        self.end = end
        self.end_error = end_error
        self.age_derivation = age_derivation
        #self.events = self.italisize(events)
        self.events = events
        self.gssp = gssp
        self.status = status
        self.publications = publications
        self.textcolor = (0,0,0)
        self.textsize = textsize
        self.textrotate = textrotate
        self.indentation = ""
        self.x = 20
        self.y = end
        self.width = 20
        self.subunits = []
        self.rectangle = None
        self.color = None
        self.setColorString(colorstring)

        self.setDimensions()

    def getHeight(self):
        return (self.begin - self.end)
    
    def italisize(self,text):
        """
        This needs to be changed to something
        more elaborate that works in SVG, maybe
        use tspan.
        """
        if (string.count(text,'_') % 2) != 0:
            return text
        first = 1
        italisized_string = " "
        for char in text:
            if char == '_' and first:
                italisized_string += "<i>"
                first = 0
            elif char == '_':
                italisized_string += "</i>"
                first = 1
            else:
                italisized_string += char
        return italisized_string
        
    def add(self,subunit):
        self.subunits.append(subunit)

    def Subunits(self):
        return self.subunits
    
    def numSubunits(self):
        return len(self.subunits)

    def imageMapTag(self):
        height = (self.begin-self.end)+1

        x1 = self.x
        y1 = self.timescale.y_offset + self.y
        x2 = self.x + self.width
        y2 = self.timescale.y_offset + self.y + height

        description_string = []
        description_string += "%s (%2.1f to %2.1f Ma)<br />" % (self.name,self.begin,self.end) 

        italics_tag_alternator = 0
        italics_tags = ['<i>','</i>']
        new_description_string = []
        for character in self.events:
            if character == "'":
                new_description_string += italics_tags[italics_tag_alternator]
                if italics_tag_alternator == 0:
                    italics_tag_alternator = 1
                else:
                    italics_tag_alternator = 0
            else:
                new_description_string += character

        new_description_string = ''.join(new_description_string)
        
        description_string += "Base: %s" % new_description_string
        description_string = ''.join(description_string)

        return "<area onmouseover=\"return escape(\'%s\')\" shape=\"rect\" coords=\"%d, %d, %d, %d\">\n" % (description_string, x1, y1, x2, y2)

    def htmlForListing(self):
        var = []
        var += "<dl>\n"
        var += "    <dt>%s: (%2.1f to %2.1f Ma)</dt>\n" % (self.name,self.begin,self.end) 
        var += "    <dd>Base: %s</dd>\n" % (self.events)
        var += "</dl>\n" 

        return var

    def rl_draw_Rectangle(self, drawing):
        height = (self.begin-self.end)+1
        drawing.add(Rect(self.x,
                         yUp(self.y+height),
                         self.width,
                         height,
                         fillColor=getrlcolor(self.color[0],self.color[1],self.color[2])))

    def strarray_Rectangle(self):
        height = (self.begin-self.end)+1

        var = ["%s<rect id=\"%s\" x=\"%d\" y=\"%d\" height=\"%d\" " %\
               (self.indentation,self.name.replace(' ','_'),self.x,self.y,height),
               "width=\"%d\" style=\"stroke-width:0.2;fill:%s;\"><title>%s (%2.1f to %2.1f Ma)</title></rect>\n" %\
               (self.width,colorstr(self.color),self.name,self.begin,self.end)] 

        return var

    def setDimensions(self):
        if self.name.endswith('Time'):
            self.indentation = "    "
            self.x = self.timescale.x_offset
            self.y = self.timescale.y_offset + self.end
            self.textrotate = -90
            self.textsize = 8
            self.width = self.timescale.geologic_time_width
        elif self.name.endswith('Eon'):
            self.indentation = "        "
            self.x = self.timescale.x_offset + self.timescale.geologic_time_width
            self.y = self.timescale.y_offset + self.end
            self.textrotate = -90
            self.textsize = 14
            self.width = self.timescale.eon_width
        elif self.name.endswith('Era'):
            self.indentation = "            "
            self.x = self.timescale.x_offset + \
                     self.timescale.geologic_time_width + \
                     self.timescale.eon_width
            self.y = self.timescale.y_offset + self.end
            self.textrotate = -90
            self.textsize = 14
            self.width = self.timescale.era_width
        elif self.name.endswith('System'):
            self.indentation = "                "
            self.x = self.timescale.x_offset + \
                     self.timescale.geologic_time_width + \
                     self.timescale.eon_width + \
                     self.timescale.era_width
            self.y = self.timescale.y_offset + self.end
            if self.timescale.simple_version:
                self.textrotate = 0
                self.textsize = 20
            else:
                self.textrotate = -90
                self.textsize = 5
            self.width = self.timescale.system_width
        elif self.name.endswith('Series'):
            self.indentation = "                    "
            self.x = self.timescale.x_offset + \
                     self.timescale.geologic_time_width + \
                     self.timescale.eon_width + \
                     self.timescale.era_width + \
                     self.timescale.system_width
            self.y = self.timescale.y_offset + self.end
            self.textrotate = 0
            self.textsize = 3
            self.width = self.timescale.series_width
        else: # stage
            self.indentation = "                        "
            self.x = self.timescale.x_offset + \
                     self.timescale.geologic_time_width + \
                     self.timescale.eon_width + \
                     self.timescale.era_width + \
                     self.timescale.system_width + \
                     self.timescale.series_width
            self.y = self.timescale.y_offset + self.end
            self.textrotate = 0
            self.textsize = 3
            self.width = self.timescale.stage_width
        if self.textrotate == 0:
            self.text_x = self.x + (self.width / 2)
            self.text_y = self.y + ((self.begin-self.end) / 2) + (self.textsize / 3)
        else:
            self.text_x = self.x + (self.width / 2) + (self.textsize / 3)
            self.text_y = self.y + ((self.begin-self.end) / 2)
        
    def updatePositioningAndDimensions(self):
        self.setDimensions()
        
    def simplePrint(self):
        print "%s%s, %2.1f,%2.1f (%2.1f), x=%2.1f, y=%2.1f" %\
              (self.indentation,self.name,self.end,self.begin,self.begin-self.end,self.x,self.y)
        for subunit in self.subunits: subunit.simplePrint()            
        
    def ColorString(self):
        return self.colorstring
    
    def getBegin(self):
        return self.begin

    def getBeginError(self):
        return self.begin_error

    def getEnd(self):
        return self.end

    def getEndError(self):
        return self.end_error

    def setEnd(self,end):
        self.end = end
        
    def setEndError(self,end_error):
        self.end_error = end_error
        
    def setColorString(self,colorstring):
        self.colorstring = colorstring
        if len(colorstring) > 0:
            rawcolorstring = self.colorstring[colorstring.index('(')+1:colorstring.index(')')]
            red,green,blue = string.split(rawcolorstring,',')
            self.color = (int(red),int(green),int(blue))
        
    def setBegin(self,begin):
        self.begin = begin
        
    def setBeginError(self,begin_error):
        self.begin_error = begin_error

    def getDuration(self):
        begin = float(self.begin)
        end = float(self.end)
        return (begin - end)
    
    def Name(self):
        return self.name

    def AgeDerivation(self):
        return self.age_derivation

    def Events(self):
        return self.events

    def GSSP(self):
        return self.gssp

    def Status(self):
        return self.status
    
    def Publications(self):
        return self.publications

    def setAgeDerivation(self,age_derivation):
        self.age_derivation = age_derivation

    def setEvents(self,events):
        self.events = events

    def setGSSP(self,gssp):
        self.gssp = gssp

    def setStatus(self,status):
        self.status = status
    
    def setPublications(self,publications):
        self.publications = publications

    def isUnit(self,name):
        if self.name == name:
            return 1
        else:
            return 0

    def strippedName(self):
        if self.name.endswith('Time'):
            return self.name
        elif self.name.endswith('Eon'):
            return self.name[0:-4]
        elif self.name.endswith('Era'):
            return self.name[0:-4]
        elif self.name.endswith('System'):
            return self.name[0:-7]
        elif self.name.endswith('SubSystem'):
            return self.name[0:-10]
        elif self.name.endswith('Series'):
            return self.name[0:-7]
        elif self.name.endswith('Subseries'):
            return self.name[0:-10]
        elif self.name.endswith('Stage'):
            return self.name[0:-6]
        else: 
            return self.name[0:-6]

    def rl_draw(self,drawing):
        self.rl_draw_Rectangle(drawing)

        text_x = self.x + (self.width / 2)
        text_y = self.y + ((self.begin-self.end) / 2)

        label = Label()
        label.setOrigin(text_x, yUp(text_y))
        label.angle = self.textrotate * -1
        label.boxAnchor = 'c'
        label.fillColor = getrlcolor(self.textcolor[0],self.textcolor[1],self.textcolor[2])
        label.fontSize = self.textsize
        label.setText(self.strippedName())

        drawing.add(label)

        for subunit in self.subunits: 
            subunit.rl_draw(drawing)

    def strarray(self):
        var = []
        var += self.strarray_Rectangle()
        var += ["%s<g transform=\"translate(%d,%d)\">\n" %\
                (self.indentation,self.text_x,self.text_y),
                "%s    <text class=\"str0\" transform=\"rotate(%d)\" style=\"font-size:%d;fill:%s;text-anchor:middle\">%s</text>\n" %\
                (self.indentation,self.textrotate,self.textsize,colorstr(self.textcolor),self.strippedName()),
                "%s</g>\n"  % self.indentation]
        for subunit in self.subunits: 
            var += subunit.strarray()            
        return var

    def getLowestTerminalUnit(self,parent):
        if self == parent.subunits[-1] and len(self.subunits) == 0:
            return self
        else:
            lowest_unit = self
            for subunit in self.subunits:
                lowest_unit = subunit.getLowestTerminalUnit(self)
            return lowest_unit

    def getTimeUnits(self,time_units,target_type):
        if self.name.endswith(target_type):
            return time_units.append(self)
        else:
            for subunit in self.subunits:
                subunit.getTimeUnits(time_units,target_type)

    def getRight(self):
        return self.x + self.width

    def getBottom(self):
        return self.y + self.getHeight()

    def rl_drawRightBoundaryLabels(self,var,target_unit_type,parent,drawing):
        if self.name.endswith(target_unit_type):
            lowest_unit = self.getLowestTerminalUnit(parent)
            label = Label()
            label.setOrigin(lowest_unit.getRight()+5, yUp(lowest_unit.getBottom()))
            label.boxAnchor = 'w'
            label.fontSize = 8
            label.setText("%s Ma" % lowest_unit.begin)
            drawing.add(label)
        else:
            for subunit in self.subunits:
                subunit.rl_drawRightBoundaryLabels(var,target_unit_type,self,drawing)

    def getRightBoundaryLabels(self,var,target_unit_type,parent):
        if self.name.endswith(target_unit_type):
            lowest_unit = self.getLowestTerminalUnit(parent)
            var += ["%s<g transform=\"translate(%d,%d)\">\n" %\
                    (lowest_unit.indentation,lowest_unit.getRight()+5,lowest_unit.getBottom()+3),
                    "%s    <text class=\"str0\" style=\"font-size:%d;fill:%s\">" %\
                    (lowest_unit.indentation,8,colorstr((0,0,0))),
                    "%s Ma</text>\n" % (lowest_unit.begin),
                    "%s</g>\n"  % lowest_unit.indentation]
            return
        else:
            for subunit in self.subunits:
                subunit.getRightBoundaryLabels(var,target_unit_type,self)
            return

class GeologicTimeScaleDrawing:
    def __init__(self, render_types, height, width, x_offset, y_offset, end_time, start_time, scale_to_fit, simple_version, point_events_input, range_events_input):
        self.render_types = render_types
        self.height = height
        self.width = width
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.end_time = end_time
        self.start_time = start_time
        self.scale_to_fit = scale_to_fit
        self.simple_version = simple_version
        self.point_events_input = point_events_input
        self.range_events_input = range_events_input

        name = "timescalefull"
        system_width = 20

        if simple_version:
            name = "timescalesimple"
            system_width = 140

        self.timescale = GeologicTimeScaleDrawingEngine(self.parsePointEvents(),
                                                        self.parseRangeEvents(),
                                                        name,
                                                        height,
                                                        width,
                                                        x_offset,
                                                        y_offset,
                                                        end_time,
                                                        start_time,
                                                        geologic_time_width = 20,
                                                        eon_width = 20,
                                                        era_width = 20,
                                                        system_width = system_width,
                                                        period_width = 100,
                                                        series_width = 50,
                                                        stage_width = 100,
                                                        simple_version = simple_version)

        if simple_version:
            self.timescale.pruneForSimple()

        self.timescale.pruneForStartAndEnd(start_time, end_time)
            
    def parsePointEvents(self):
        point_events = []
        for event_string in self.point_events_input:
            label_placement,label_angle,x_position,symbol_type,symbol_width,symbol_height,time,font_size,text,red,green,blue = eval('[%s]' % event_string)
            point_events.append(PointEvent(label_placement,int(label_angle),int(x_position),symbol_type,int(symbol_width),int(symbol_height),int(time),font_size,text,(int(red),int(green),int(blue))))
        return point_events

    def parseRangeEvents(self):
        range_events = []
        for event_string in self.range_events_input:
            label_placement,label_angle,x_position,bar_width,event_start_time,event_end_time,font_size,text,red,green,blue = eval('[%s]' % event_string)
            range_events.append(RangeEvent(label_placement,int(label_angle),int(x_position),int(bar_width),int(event_start_time),int(event_end_time),font_size,text,(int(red),int(green),int(blue))))
        return range_events

    def getDrawing(self):
        self.drawing = Drawing(self.width, self.height)

        self.timescale.rl_draw(self.drawing)
        
        self.timescale.rl_drawPointEvents(self.drawing)
        
        self.timescale.rl_drawRangeEvents(self.drawing)

        if self.scale_to_fit:
            scale_factor = float(self.height) / 4567.2
            self.drawing.scale(scale_factor,scale_factor)
        else:
            # move origin down, because image 0,0 is bottom left of images (PNG, etc.)
            self.drawing.translate(0.0,-((self.y_offset+4567.2)-self.height))

        return self.drawing

    def getSVG(self):
        return self.timescale.getSVG()

    def getHTMLForListing(self):
        return self.timescale.getHTMLForListing()

    def getHTMLForContainingSVG(self, name):
        return self.timescale.getHTMLForContainingSVG(name)

    def getHTMLForContainingImageMap(self, name):
        return self.timescale.getHTMLForContainingImageMap(name)
